"""
Model metamers.

Classes to perform the synthesis of model metamers.
"""

import re
import warnings
from collections import OrderedDict
from collections.abc import Callable
from typing import Any, Literal

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import torch
from torch import Tensor
from tqdm.auto import tqdm

from ..tools import data, display, optim, signal
from ..tools.convergence import coarse_to_fine_enough, loss_convergence
from ..tools.validate import validate_coarse_to_fine, validate_input, validate_model
from .synthesis import OptimizedSynthesis


class Metamer(OptimizedSynthesis):
    r"""
    Synthesize metamers for image-computable differentiable models.

    Following the basic idea in [1]_, this class creates a metamer for a given model on
    a given image. We iteratively adjust the pixel values so as to match the
    representation of the :attr:`metamer` and :attr:`image`.

    Parameters
    ----------
    image
        A tensor, this is the image whose representation we wish to
        match.
    model
        A visual model.
    loss_function
        The loss function to use to compare the representations of the models
        in order to determine their loss.
    range_penalty_lambda
        Strength of the regularizer that enforces the allowed_range. Must be
        non-negative.
    allowed_range
        Range (inclusive) of allowed pixel values. Any values outside this
        range will be penalized.

    References
    ----------
    .. [1] J Portilla and E P Simoncelli. A Parametric Texture Model
       based on Joint Statistics of Complex Wavelet Coefficients. Int'l
       Journal of Computer Vision. 40(1):49-71, October, 2000.
       https://www.cns.nyu.edu/~eero/ABSTRACTS/portilla99-abstract.html
       https://www.cns.nyu.edu/~lcv/texture/
    """

    def __init__(
        self,
        image: Tensor,
        model: torch.nn.Module,
        loss_function: Callable[[Tensor, Tensor], Tensor] = optim.mse,
        range_penalty_lambda: float = 0.1,
        allowed_range: tuple[float, float] = (0, 1),
    ):
        super().__init__(range_penalty_lambda, allowed_range)
        validate_input(image, allowed_range=allowed_range)
        validate_model(
            model,
            image_shape=image.shape,
            image_dtype=image.dtype,
            device=image.device,
        )
        self._model = model
        self._image = image
        self._image_shape = image.shape
        self._target_representation = self.model(self.image)
        self._scheduler = None
        self._scheduler_step_arg = False
        self.loss_function = loss_function
        self._saved_metamer = []
        self._store_progress = None
        self._metamer = None

    def setup(
        self,
        initial_image: Tensor | None = None,
        optimizer: torch.optim.Optimizer | None = None,
        optimizer_kwargs: dict | None = None,
        scheduler: torch.optim.lr_scheduler.LRScheduler | None = None,
        scheduler_kwargs: dict | None = None,
    ):
        """
        Initialize the metamer, optimizer, and scheduler.

        Can only be called once. If ``load()`` has been called, ``initial_image`` must
        be ``None``.

        Parameters
        ----------
        initial_image
            The tensor we use to initialize the metamer. If ``None``, we initialize with
            uniformly-distributed random noise lying within ``self.allowed_range``.
        optimizer
            The un-initialized optimizer object to use. If ``None``, we use
            :class:`torch.optim.Adam`.
        optimizer_kwargs
            The keyword arguments to pass to the optimizer on initialization. If
            ``None``, we use ``{"lr": .01}`` and, if optimizer is ``None``,
            ``{"amsgrad": True}``.
        scheduler
            The un-initialized learning rate scheduler object to use. If ``None``, we
            don't use one.
        scheduler_kwargs
            The keyword arguments to pass to the scheduler on initialization.

        Raises
        ------
        ValueError
            If you try to set ``initial_image`` after calling :func:`load`.
        ValueError
            If ``setup`` is called more than once or after :func:`synthesize`.
        ValueError
            If you try to set ``optimizer_kwargs`` after calling :func:`load`.
        TypeError
            If the loaded object had a non-Adam optimizer, but the ``optimizer`` arg
            is not specified.
        ValueError
            If the loaded object had an optimizer, and the ``optimizer`` arg is
            a different type.
        ValueError
            If you try to set ``scheduler_kwargs`` after calling :func:`load`.
        TypeError
            If the loaded object had a scheduler, but the ``scheduler`` arg is not
            specified.
        ValueError
            If the loaded object had a scheduler, but the ``scheduler`` arg is
            a different type.

        Warns
        -----
        UserWarning
            If ``initial_image`` is a different shape than ``self.image``.

        Examples
        --------
        Set initial image:

        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> model = po.simul.Gaussian(30).eval()
        >>> po.tools.remove_grad(model)
        >>> met = po.synth.Metamer(img, model)
        >>> met.setup(po.data.curie())
        >>> met.synthesize(10)

        Set optimizer:

        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> model = po.simul.Gaussian(30).eval()
        >>> po.tools.remove_grad(model)
        >>> met = po.synth.Metamer(img, model)
        >>> met.setup(optimizer=torch.optim.SGD, optimizer_kwargs={"lr": 0.01})
        >>> met.synthesize(10)

        Use with save/load. Only the optimizer object is necessary, its kwargs and the
        initial image are handled by load.

        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> model = po.simul.Gaussian(30).eval()
        >>> po.tools.remove_grad(model)
        >>> met = po.synth.Metamer(img, model)
        >>> met.setup(
        ...     po.data.curie(),
        ...     optimizer=torch.optim.SGD,
        ...     optimizer_kwargs={"lr": 0.01},
        ... )
        >>> met.synthesize(10)
        >>> met.save("metamer_setup.pt")
        >>> met = po.synth.Metamer(img, model)
        >>> met.load("metamer_setup.pt")
        >>> met.setup(optimizer=torch.optim.SGD)
        >>> met.synthesize(10)
        """
        if self._metamer is None:
            if initial_image is None:
                metamer = torch.rand_like(self.image)
                # rescale metamer to lie within the interval
                # self.allowed_range
                metamer = signal.rescale(metamer, *self.allowed_range)
            else:
                validate_input(initial_image, allowed_range=self.allowed_range)
                if initial_image.size() != self.image.size():
                    warnings.warn(
                        "initial_image and image are different sizes! This "
                        "has not been tested as much, open an issue if you have "
                        "any problems! https://github.com/plenoptic-org/plenoptic/"
                        "issues/new?template=bug_report.md"
                    )
                metamer = initial_image.clone().detach()
                metamer = metamer.to(dtype=self.image.dtype, device=self.image.device)
            metamer.requires_grad_()
            self._metamer = metamer
        else:
            if self._loaded:
                if initial_image is not None:
                    raise ValueError("Cannot set initial_image after calling load()!")
            else:
                raise ValueError(
                    "setup() can only be called once and must be called"
                    " before synthesize()!"
                )

        # initialize the optimizer
        self._initialize_optimizer(optimizer, self.metamer, optimizer_kwargs)
        # and scheduler
        self._initialize_scheduler(scheduler, self.optimizer, scheduler_kwargs)
        # reset _loaded, if everything ran successfully
        self._loaded = False

    def synthesize(
        self,
        max_iter: int = 100,
        store_progress: bool | int = False,
        stop_criterion: float = 1e-4,
        stop_iters_to_check: int = 50,
    ):
        r"""
        Synthesize a metamer.

        Update the pixels of :attr:`metamer` until its representation matches that of
        :attr:`image`.

        We run this until either we reach ``max_iter`` or the change over the
        past ``stop_iters_to_check`` iterations is less than
        ``stop_criterion``, whichever comes first.

        Parameters
        ----------
        max_iter
            The maximum number of iterations to run before we end synthesis
            (unless we hit the stop criterion).
        store_progress
            Whether we should store the metamer image in progress during
            synthesis. If ``False``, we don't save anything. If True, we save every
            iteration. If an int, we save every ``store_progress`` iterations
            (note then that 0 is the same as False and 1 the same as True).
        stop_criterion
            If the loss over the past ``stop_iters_to_check`` has changed
            less than ``stop_criterion``, we terminate synthesis.
        stop_iters_to_check
            How many iterations back to check in order to see if the
            loss has stopped decreasing (for ``stop_criterion``).

        Raises
        ------
        ValueError
            If we find a NaN during optimization.
        """
        # if setup hasn't been called manually, call it now.
        if self._metamer is None or isinstance(self._scheduler, tuple):
            self.setup()
        self._current_loss = None

        # get ready to store progress
        self.store_progress = store_progress

        pbar = tqdm(range(max_iter))

        for i in pbar:
            # update saved_* attrs. len(_losses) gives the total number of
            # iterations and will be correct across calls to `synthesize`
            self._store(len(self._losses))

            loss = self._optimizer_step(pbar)

            if not np.isfinite(loss):
                raise ValueError("Found a NaN in loss during optimization.")

            if self._check_convergence(stop_criterion, stop_iters_to_check):
                warnings.warn("Loss has converged, stopping synthesis")
                break

        # compute current loss, no need to compute gradient
        with torch.no_grad():
            self._current_loss = self.objective_function().item()

        pbar.close()

    def objective_function(
        self,
        metamer: Tensor | None = None,
        target_representation: Tensor | None = None,
        **analyze_kwargs: Any,
    ) -> Tensor:
        """
        Compute the metamer synthesis loss.

        This calls self.loss_function on
        ``self.model(metamer, **analyze_kwargs)`` and
        ``target_representation`` and then adds the weighted range penalty
        on ``metamer``.

        Parameters
        ----------
        metamer
            Current ``metamer``. If ``None``, we use ``self.metamer``.
        target_representation
            Model response to ``image``. If ``None``, we use
            ``self.target_representation``.
        **analyze_kwargs
            Additional kwargs to pass to ``self.model(metamer)``.

        Returns
        -------
        loss
            1-element tensor containing the loss on this step.
        """
        if metamer is None:
            metamer = self.metamer
            # if this is empty, then self.metamer hasn't been initialized
            if metamer.numel() == 0:
                return torch.empty(0)
        if target_representation is None:
            target_representation = self.target_representation
        metamer_representation = self.model(metamer, **analyze_kwargs)
        loss = self.loss_function(metamer_representation, target_representation)
        range_penalty = optim.penalize_range(metamer, self.allowed_range)
        return loss + self.range_penalty_lambda * range_penalty

    def get_progress(
        self,
        iteration: int | None,
        iteration_selection: Literal["floor", "ceiling", "round"] = "round",
    ) -> dict:
        """
        Return dictionary summarizing synthesis progress at ``iteration``.

        This returns a dictionary containing info from :attr:`losses`,
        :attr:`pixel_change_norm`, :attr:`gradient_norm`, and
        :attr:`saved_metamer` corresponding to ``iteration``. If synthesis was
        run with ``store_progress=False`` (and so we did not cache anything in
        :attr:`saved_metamer`), then that key will be missing. If synthesis was
        run with ``store_progress>1``, we will grab the corresponding tensor
        from :attr:`saved_metamer`, with behavior determined by
        ``iteration_selection``.

        The returned dictionary will additionally contain the keys:

        - ``"iteration"``: the (0-indexed positive) synthesis iteration that the
          values for :attr:`losses`, :attr:`pixel_change_norm`, and
          :attr:`gradient_norm` come from.

        - If ``self.store_progress``, ``"store_progress_iteration"``: the (0-indexed
          positive) synthesis iteration that the value for :attr:`saved_metamer` comes
          from.

        Note that for the most recent iteration (``iteration=-1`` or ``iteration=None``
        or ``iteration==len(self.losses)-1``), we do not have values for
        :attr:`pixel_change_norm` or :attr:`gradient_norm`, since in this case we are
        showing the loss and value for the current metamer.

        Parameters
        ----------
        iteration
            Synthesis iteration to summarize. If ``None``, grab the most recent.
            Negative values are allowed.
        iteration_selection

            How to select the relevant iteration from :attr:`saved_metamer`
            when the request iteration wasn't stored.

            When synthesis was run with ``store_progress=n`` (where ``n>1``),
            metamers are only saved every ``n`` iterations. If you request an
            iteration where a metamer wasn't saved, this determines which available
            iteration is used instead:

            * ``"floor"``: use the closest saved iteration **before** the
              requested one.

            * ``"ceiling"``: use the closest saved iteration **after** the
              requested one.

            * ``"round"``: use the closest saved iteration.

        Returns
        -------
        progress_info
            Dictionary summarizing synthesis progress.

        Raises
        ------
        IndexError
            If ``iteration`` takes an illegal value.

        Warns
        -----
        UserWarning
            If the iteration used for ``saved_metamer`` is not the same as the argument
            ``iteration`` (because e.g., you set ``iteration=3`` but
            ``self.store_progress=2``).
        """
        return super().get_progress(
            iteration,
            iteration_selection,
            store_progress_attributes=["saved_metamer"],
        )

    def _optimizer_step(self, pbar: tqdm) -> Tensor:
        r"""
        Compute and propagate gradients, then step the optimizer to update metamer.

        Parameters
        ----------
        pbar
            A tqdm progress-bar, which we update with a postfix
            describing the current loss, gradient norm, and learning
            rate (it already tells us which iteration and the time
            elapsed).

        Returns
        -------
        loss
            1-element tensor containing the loss on this step.
        """  # numpydoc ignore=ES01
        last_iter_metamer = self.metamer.clone()
        loss = self.optimizer.step(self._closure)
        self._losses.append(loss)

        grad_norm = torch.linalg.vector_norm(self.metamer.grad.data, ord=2, dim=None)
        self._gradient_norm.append(grad_norm.item())

        # optionally step the scheduler, passing loss if needed
        if self.scheduler is not None:
            if self._scheduler_step_arg:
                self.scheduler.step(loss)
            else:
                self.scheduler.step()

        pixel_change_norm = torch.linalg.vector_norm(
            self.metamer - last_iter_metamer, ord=2, dim=None
        )
        self._pixel_change_norm.append(pixel_change_norm.item())
        # add extra info here if you want it to show up in progress bar
        pbar.set_postfix(
            OrderedDict(
                loss=f"{loss:.04e}",
                learning_rate=self.optimizer.param_groups[0]["lr"],
                gradient_norm=f"{grad_norm.item():.04e}",
                pixel_change_norm=f"{pixel_change_norm.item():.04e}",
            )
        )
        return loss

    def _check_convergence(
        self, stop_criterion: float, stop_iters_to_check: int
    ) -> bool:
        r"""
        Check whether the loss has stabilized and, if so, return True.

        Uses :func:`loss_convergence`.

        Parameters
        ----------
        stop_criterion
            If the loss over the past ``stop_iters_to_check`` has changed
            less than ``stop_criterion``, we terminate synthesis.
        stop_iters_to_check
            How many iterations back to check in order to see if the
            loss has stopped decreasing (for ``stop_criterion``).

        Returns
        -------
        loss_stabilized
            Whether the loss has stabilized or not.
        """
        return loss_convergence(self, stop_criterion, stop_iters_to_check)

    def _store(self, i: int) -> bool:
        """
        Store metamer, if appropriate.

        If it's the right iteration, we update :attr:`saved_metamer`.

        Parameters
        ----------
        i
            The current iteration.

        Returns
        -------
        stored
            True if we stored this iteration, False if not.
        """
        if self.store_progress and (i % self.store_progress == 0):
            # want these to always be on cpu, to reduce memory use for GPUs
            self._saved_metamer.append(self.metamer.clone().to("cpu"))
            stored = True
        else:
            stored = False
        return stored

    def save(self, file_path: str):
        r"""
        Save all relevant variables in .pt file.

        Note that if ``store_progress`` is True, this will probably be very
        large.

        See :func:`load` docstring for an example of use.

        Parameters
        ----------
        file_path :
            The path to save the metamer object to.
        """
        save_io_attrs = [
            ("loss_function", ("_image", "_metamer")),
            ("_model", ("_image",)),
        ]
        save_state_dict_attrs = ["_optimizer", "_scheduler"]
        super().save(file_path, save_io_attrs, save_state_dict_attrs)

    def to(self, *args: Any, **kwargs: Any):
        r"""
        Move and/or cast the parameters and buffers.

        This can be called as

        .. function:: to(device=None, dtype=None, non_blocking=False)

        .. function:: to(dtype, non_blocking=False)

        .. function:: to(tensor, non_blocking=False)

        Its signature is similar to :meth:`torch.Tensor.to`, but only accepts
        floating point desired :attr:`dtype` s. In addition, this method will
        only cast the floating point parameters and buffers to :attr:`dtype`
        (if given). The integral parameters and buffers will be moved
        :attr:`device`, if that is given, but with dtypes unchanged. When
        :attr:`non_blocking` is set, it tries to convert/move asynchronously
        with respect to the host if possible, e.g., moving CPU Tensors with
        pinned memory to CUDA devices.

        See :meth:`torch.nn.Module.to` for examples.

        .. note::
            This method modifies the module in-place.

        Parameters
        ----------
        device : torch.device
            The desired device of the parameters and buffers in this module.
        dtype : torch.dtype
            The desired floating point type of the floating point parameters and
            buffers in this module.
        tensor : torch.Tensor
            Tensor whose dtype and device are the desired dtype and device for
            all parameters and buffers in this module.
        """  # numpydoc ignore=PR01,PR02
        attrs = ["_image", "_target_representation", "_metamer", "_saved_metamer"]
        super().to(*args, attrs=attrs, **kwargs)
        # try to call .to() on model. this should work, but it might fail if e.g., this
        # a custom model that doesn't inherit torch.nn.Module
        try:
            self._model = self._model.to(*args, **kwargs)
        except AttributeError:
            warnings.warn("Unable to call model.to(), so we leave it as is.")

    def load(
        self,
        file_path: str,
        map_location: str | None = None,
        tensor_equality_atol: float = 1e-8,
        tensor_equality_rtol: float = 1e-5,
        **pickle_load_args: Any,
    ):
        r"""
        Load all relevant stuff from a .pt file.

        This must be called by a ``Metamer`` object initialized just like the saved
        object.

        Note this operates in place and so doesn't return anything.

        .. versionchanged:: 1.2
           load behavior changed in a backwards-incompatible manner in order to
           compatible with breaking changes in torch 2.6.

        Parameters
        ----------
        file_path
            The path to load the synthesis object from.
        map_location
            Argument to pass to :func:`torch.load` as ``map_location``. If you
            save stuff that was being run on a GPU and are loading onto a
            CPU, you'll need this to make sure everything lines up
            properly. This should be structured like the str you would
            pass to :class:`torch.device`.
        tensor_equality_atol
            Absolute tolerance to use when checking for tensor equality during load,
            passed to :func:`torch.allclose`. It may be necessary to increase if you are
            saving and loading on two machines with torch built by different cuda
            versions. Be careful when changing this! See
            :class:`torch.finfo<torch.torch.finfo>` for more details about floating
            point precision of different data types (especially, ``eps``); if you have
            to increase this by more than 1 or 2 decades, then you are probably not
            dealing with a numerical issue.
        tensor_equality_rtol
            Relative tolerance to use when checking for tensor equality during load,
            passed to :func:`torch.allclose`. It may be necessary to increase if you are
            saving and loading on two machines with torch built by different cuda
            versions. Be careful when changing this! See
            :class:`torch.finfo<torch.torch.finfo>` for more details about floating
            point precision of different data types (especially, ``eps``); if you have
            to increase this by more than 1 or 2 decades, then you are probably not
            dealing with a numerical issue.
        **pickle_load_args
            Any additional kwargs will be added to ``pickle_module.load`` via
            :func:`torch.load`, see that function's docstring for details.

        Raises
        ------
        ValueError
            If :func:`setup` or :func:`synthesize` has been called before this call
            to ``load``.
        ValueError
            If the object saved at ``file_path`` is not a ``Metamer`` object.
        ValueError
            If the saved and loading ``Metamer`` objects have a different value
            for any of :attr:`image`, :attr:`range_penalty_lambda`,
            or :attr:`allowed_range`.
        ValueError
            If the behavior of :attr:`loss_function` or :attr:`model` is different
            between the saved and loading objects.

        Warns
        -----
        UserWarning
            If :func:`setup` will need to be called after load, to finish initializing
            :attr:`optimizer` or :attr:`scheduler`.

        See Also
        --------
        :func:`~plenoptic.tools.io.examine_saved_synthesis`
            Examine metadata from saved object: pytorch and plenoptic versions, name of
            the synthesis object, shapes of tensors, etc.

        Examples
        --------
        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> model = po.simul.Gaussian(30).eval()
        >>> po.tools.remove_grad(model)
        >>> metamer = po.synth.Metamer(img, model)
        >>> metamer.synthesize(max_iter=5, store_progress=True)
        >>> metamer.save("metamers.pt")
        >>> metamer_copy = po.synth.Metamer(img, model)
        >>> metamer_copy.load("metamers.pt")
        """
        self._load(
            file_path,
            map_location,
            tensor_equality_atol=tensor_equality_atol,
            tensor_equality_rtol=tensor_equality_rtol,
            **pickle_load_args,
        )

    def _load(
        self,
        file_path: str,
        map_location: str | None = None,
        additional_check_attributes: list[str] = [],
        additional_check_io_attributes: list[str] = [],
        tensor_equality_atol: float = 1e-8,
        tensor_equality_rtol: float = 1e-5,
        **pickle_load_args: Any,
    ):
        r"""
        Load from a file.

        This is a helper function for loading.

        Users interact with ``load`` (without the underscore), this is to allow
        subclasses to specify additional attributes or loss functions to check.

        Parameters
        ----------
        file_path
            The path to load the synthesis object from.
        map_location
            Argument to pass to :func:`torch.load` as ``map_location``. If you
            save stuff that was being run on a GPU and are loading onto a
            CPU, you'll need this to make sure everything lines up
            properly. This should be structured like the str you would
            pass to :class:`torch.device`.
        additional_check_attributes
            Any additional attributes to check for equality. Intended for use by any
            subclasses, to add other attributes set at initialization.
        additional_check_io_attributes
            Any additional attributes whose input/output behavior we should check.
            Intended for use by any subclasses.
        tensor_equality_atol
            Absolute tolerance to use when checking for tensor equality during load,
            passed to :func:`torch.allclose`. It may be necessary to increase if you are
            saving and loading on two machines with torch built by different cuda
            versions. Be careful when changing this! See
            :class:`torch.finfo<torch.torch.finfo>` for more details about floating
            point precision of different data types (especially, ``eps``); if you have
            to increase this by more than 1 or 2 decades, then you are probably not
            dealing with a numerical issue.
        tensor_equality_rtol
            Relative tolerance to use when checking for tensor equality during load,
            passed to :func:`torch.allclose`. It may be necessary to increase if you are
            saving and loading on two machines with torch built by different cuda
            versions. Be careful when changing this! See
            :class:`torch.finfo<torch.torch.finfo>` for more details about floating
            point precision of different data types (especially, ``eps``); if you have
            to increase this by more than 1 or 2 decades, then you are probably not
            dealing with a numerical issue.
        **pickle_load_args
            Any additional kwargs will be added to ``pickle_module.load`` via
            :func:`torch.load`, see that function's docstring for details.
        """
        check_attributes = [
            "_image",
            "_range_penalty_lambda",
            "_allowed_range",
        ]
        check_attributes += additional_check_attributes
        check_io_attrs = [
            ("loss_function", ("_image", "_metamer")),
            ("_model", ("_image",)),
        ]
        check_io_attrs += additional_check_io_attributes
        super().load(
            file_path,
            "_metamer",
            map_location=map_location,
            check_attributes=check_attributes,
            check_io_attributes=check_io_attrs,
            state_dict_attributes=["_optimizer", "_scheduler"],
            tensor_equality_atol=tensor_equality_atol,
            tensor_equality_rtol=tensor_equality_rtol,
            **pickle_load_args,
        )
        # make this require a grad again
        self.metamer.requires_grad_()
        # these are always supposed to be on cpu, but may get copied over to
        # gpu on load (which can cause problems when resuming synthesis), so
        # fix that.
        if len(self._saved_metamer) and self._saved_metamer[0].device.type != "cpu":
            self._saved_metamer = [met.to("cpu") for met in self._saved_metamer]

    @property
    def model(self) -> torch.nn.Module:
        """The model for which the metamer is synthesized."""
        # numpydoc ignore=RT01,ES01
        return self._model

    @property
    def image(self) -> torch.Tensor:
        """Target image of metamer optimization."""
        # numpydoc ignore=RT01,ES01
        return self._image

    @property
    def target_representation(self) -> torch.Tensor:
        """
        :attr:`model` representation of :attr:`image`, i.e. ``model(image)``.

        The goal of synthesis is for ``model(metamer)`` to match this value.
        """  # numpydoc ignore=RT01
        return self._target_representation

    @property
    def metamer(self) -> torch.Tensor:
        """Model metamer, the parameter we are optimizing."""
        # numpydoc ignore=RT01,ES01
        if self._metamer is None:
            return torch.empty(0)
        return self._metamer

    @property
    def saved_metamer(self) -> torch.Tensor:
        """
        :attr:`metamer`, cached over time for later examination.

        How often the metamer is cached is determined by the ``store_progress`` argument
        to the :func:`synthesize` function.

        The last entry will always be the current :attr:`metamer`.

        If ``store_progress==1``, then this corresponds directly to :attr:`losses`:
        ``losses[i]`` is the error for ``saved_metamer[i]``

        This tensor always lives on the CPU, regardless of the device of the ``Metamer``
        object.
        """  # numpydoc ignore=RT01
        if self._metamer is None:
            return torch.empty(0)
        else:
            # for memory purposes, always on CPU
            return torch.stack([*self._saved_metamer, self.metamer.to("cpu")])


class MetamerCTF(Metamer):
    """
    Synthesize model metamers with coarse-to-fine synthesis.

    This is a special case of ``Metamer``, which uses the coarse-to-fine
    synthesis procedure described in [1]_: we start by updating metamer with
    respect to only a subset of the model's representation (generally, that
    which corresponds to the lowest spatial frequencies), and changing which
    subset we consider over the course of synthesis. This is similar to
    optimizing with a blurred version of the objective function and gradually
    adding in finer details. It improves synthesis performance for some models.

    Parameters
    ----------
    image
        A tensor, this is the image whose representation we wish to
        match.
    model
        A visual model.
    loss_function
        The loss function to use to compare the representations of the models
        in order to determine their loss.
    range_penalty_lambda
        Strength of the regularizer that enforces the allowed_range. Must be
        non-negative.
    allowed_range
        Range (inclusive) of allowed pixel values. Any values outside this
        range will be penalized.
    coarse_to_fine
        - ``"together"``: start with the coarsest scale, then gradually
          add each finer scale.
        - ``"separate"``: compute the gradient with respect to each
          scale separately (ignoring the others), then with respect
          to all of them at the end.
        (see :ref:`Metamer tutorial <metamer-nb>`_ for more details).

    References
    ----------
    .. [1] J Portilla and E P Simoncelli. A Parametric Texture Model
       based on Joint Statistics of Complex Wavelet Coefficients. Int'l
       Journal of Computer Vision. 40(1):49-71, October, 2000.
       https://www.cns.nyu.edu/~eero/ABSTRACTS/portilla99-abstract.html
       https://www.cns.nyu.edu/~lcv/texture/
    """

    def __init__(
        self,
        image: Tensor,
        model: torch.nn.Module,
        loss_function: Callable[[Tensor, Tensor], Tensor] = optim.mse,
        range_penalty_lambda: float = 0.1,
        allowed_range: tuple[float, float] = (0, 1),
        coarse_to_fine: Literal["together", "separate"] = "together",
    ):
        super().__init__(
            image,
            model,
            loss_function,
            range_penalty_lambda,
            allowed_range,
        )
        self._init_ctf(coarse_to_fine)

    def _init_ctf(self, coarse_to_fine: Literal["together", "separate"]):
        """
        Initialize stuff related to coarse-to-fine.

        - Validates value of ``coarse_to_fine``

        - Validates ``self.model`` for coarse-to-fine synthesis (calls
          :func:`validate_coarse_to_fine`).

        - Initializes attributes for coarse-to-fine synthesis.

        Parameters
        ----------
        coarse_to_fine
            Which mode of coarse-to-fine to use, see initial docstring for details.

        Raises
        ------
        ValueError
            If ``coarse_to_fine`` takes an illegal value.
        """
        # this will hold the reduced representation of the target image.
        if coarse_to_fine not in ["separate", "together"]:
            raise ValueError(
                f"Don't know how to handle value {coarse_to_fine}!"
                " Must be one of: 'separate', 'together'"
            )
        self._ctf_target_representation = None
        validate_coarse_to_fine(
            self.model, image_shape=self.image.shape, device=self.image.device
        )
        # if self.scales is not None, we're continuing a previous version
        # and want to continue. this list comprehension creates a new
        # object, so we don't modify model.scales
        self._scales = [i for i in self.model.scales[:-1]]
        if coarse_to_fine == "separate":
            self._scales += [self.model.scales[-1]]
        self._scales += ["all"]
        self._scales_timing = dict((k, []) for k in self.scales)
        self._scales_timing[self.scales[0]].append(0)
        self._scales_loss = []
        self._scales_finished = []
        self._coarse_to_fine = coarse_to_fine
        self._initial_lr = None

    def _initialize_optimizer(
        self,
        optimizer: torch.optim.Optimizer | None,
        synth_attr: torch.Tensor,
        optimizer_kwargs: dict | None = None,
    ):
        """
        Initialize optimizer.

        Calls ``super._initialize_optimizer()``, passing all arguments through, and also
        caches the initial learning rate (``self._initial_lr``), which we use when
        switching scales.

        Parameters
        ----------
        optimizer
            The (un-initialized) optimizer object to use. If ``None``, we use
            :class:`torch.optim.Adam`.
        synth_attr
            The tensor we will optimize.
        optimizer_kwargs
            The keyword arguments to pass to the optimizer on initialization. If
            ``None``, we use ``{"lr": .01}`` and, if optimizer is ``None``,
            ``{"amsgrad": True}``.
        """
        super()._initialize_optimizer(optimizer, synth_attr, optimizer_kwargs)
        # save the initial learning rate so we can reset it when we change scales
        self._initial_lr = [pg["lr"] for pg in self.optimizer.param_groups]

    def synthesize(
        self,
        max_iter: int = 100,
        store_progress: bool | int = False,
        stop_criterion: float = 1e-4,
        stop_iters_to_check: int = 50,
        change_scale_criterion: float | None = 1e-2,
        ctf_iters_to_check: int = 50,
    ):
        r"""
        Synthesize a metamer.

        Update the pixels of ``metamer`` until its representation matches
        that of ``image``.

        We run this until either we reach ``max_iter`` or the change over the
        past ``stop_iters_to_check`` iterations is less than
        ``stop_criterion``, whichever comes first.

        Parameters
        ----------
        max_iter
            The maximum number of iterations to run before we end synthesis
            (unless we hit the stop criterion).
        store_progress
            Whether we should store the metamer image in progress on every
            iteration. If ``False``, we don't save anything. If True, we save every
            iteration. If an int, we save every ``store_progress`` iterations
            (note then that 0 is the same as False and 1 the same as True).
        stop_criterion
            If the loss over the past ``stop_iters_to_check`` has changed
            less than ``stop_criterion``, we terminate synthesis.
        stop_iters_to_check
            How many iterations back to check in order to see if the
            loss has stopped decreasing (for ``stop_criterion``).
        change_scale_criterion
            Scale-specific analogue of ``change_scale_criterion``: we consider
            a given scale finished (and move onto the next) if the loss has
            changed less than this in the past ``ctf_iters_to_check``
            iterations. If ``None``, we'll change scales as soon as we've spent
            ``ctf_iters_to_check`` on a given scale.
        ctf_iters_to_check
            Scale-specific analogue of ``stop_iters_to_check``: how many
            iterations back in order to check in order to see if we should
            switch scales.

        Raises
        ------
        ValueError
            If ``stop_criterion >= change_scale_criterion`` -- behavior is strange
            otherwise.
        ValueError
            If we find a NaN during optimization.
        """
        if (change_scale_criterion is not None) and (
            stop_criterion >= change_scale_criterion
        ):
            raise ValueError(
                "stop_criterion must be strictly less than "
                "change_scale_criterion, or things get weird!"
            )

        # if setup hasn't been called manually, call it now.
        if self._metamer is None or isinstance(self._scheduler, tuple):
            self.setup()
        self._current_loss = None

        # get ready to store progress
        self.store_progress = store_progress

        pbar = tqdm(range(max_iter))

        for i in pbar:
            # update saved_* attrs. len(_losses) gives the total number of
            # iterations and will be correct across calls to `synthesize`
            self._store(len(self._losses))

            loss = self._optimizer_step(
                pbar, change_scale_criterion, ctf_iters_to_check
            )

            if not np.isfinite(loss):
                raise ValueError("Found a NaN in loss during optimization.")

            if self._check_convergence(
                i, stop_criterion, stop_iters_to_check, ctf_iters_to_check
            ):
                warnings.warn("Loss has converged, stopping synthesis")
                break

        # compute current loss, no need to compute gradient
        with torch.no_grad():
            self._current_loss = self.objective_function().item()

        pbar.close()

    def _optimizer_step(
        self,
        pbar: tqdm,
        change_scale_criterion: float,
        ctf_iters_to_check: int,
    ) -> Tensor:
        r"""
        Compute and propagate gradients, then step the optimizer to update metamer.

        Parameters
        ----------
        pbar
            A tqdm progress-bar, which we update with a postfix
            describing the current loss, gradient norm, and learning
            rate (it already tells us which iteration and the time
            elapsed).
        change_scale_criterion
            How many iterations back to check to see if the loss has stopped
            decreasing and we should thus move to the next scale in
            coarse-to-fine optimization.
        ctf_iters_to_check
            Minimum number of iterations coarse-to-fine must run at each scale.

        Returns
        -------
        loss
            1-element tensor containing the loss on this step.
        """  # numpydoc ignore=ES01
        last_iter_metamer = self.metamer.clone()

        # Check if conditions hold for switching scales:
        # - Check if loss has decreased below the change_scale_criterion and
        # - if we've been optimizing this scale for the required number of iterations
        # - The first check here is because the last scale will be 'all', and
        #   we never remove it

        if (
            len(self.scales) > 1
            and len(self.scales_loss) >= ctf_iters_to_check
            and (
                change_scale_criterion is None
                or abs(self.scales_loss[-1] - self.scales_loss[-ctf_iters_to_check])
                < change_scale_criterion
            )
            and (
                len(self._losses) - self.scales_timing[self.scales[0]][0]
                >= ctf_iters_to_check
            )
        ):
            self._scales_timing[self.scales[0]].append(len(self._losses) - 1)
            self._scales_finished.append(self._scales.pop(0))

            # Only append if scales list is still non-empty after the pop
            if self.scales:
                self._scales_timing[self.scales[0]].append(len(self._losses))

            # Reset optimizer's learning rate
            for pg, lr in zip(self.optimizer.param_groups, self._initial_lr):
                pg["lr"] = lr

            # Reset ctf target representation for the next update
            self._ctf_target_representation = None

        # the loss returned by objective_function is from *before* updating the metamer,
        # so to compute the equivalent for display purposes, we need to call this before
        # calling step()
        if self.scales[0] != "all":
            with torch.no_grad():
                overall_loss = self.objective_function(None, None).item()

        loss = self.optimizer.step(self._closure)
        if self.scales[0] == "all":
            # then the loss computed above includes all scales
            overall_loss = loss

        self._scales_loss.append(loss)
        self._losses.append(overall_loss)

        grad_norm = torch.linalg.vector_norm(self.metamer.grad.data, ord=2, dim=None)
        self._gradient_norm.append(grad_norm.item())

        # optionally step the scheduler, passing loss if needed
        if self.scheduler is not None:
            if self._scheduler_step_arg:
                self.scheduler.step(loss)
            else:
                self.scheduler.step()

        pixel_change_norm = torch.linalg.vector_norm(
            self.metamer - last_iter_metamer, ord=2, dim=None
        )
        self._pixel_change_norm.append(pixel_change_norm.item())
        # add extra info here if you want it to show up in progress bar
        pbar.set_postfix(
            OrderedDict(
                loss=f"{overall_loss:.04e}",
                learning_rate=self.optimizer.param_groups[0]["lr"],
                gradient_norm=f"{grad_norm.item():.04e}",
                pixel_change_norm=f"{pixel_change_norm.item():.04e}",
                current_scale=self.scales[0],
                current_scale_loss=f"{loss:.04e}",
            )
        )
        return overall_loss

    def _closure(self) -> Tensor:
        r"""
        Calculate the gradient, before the optimization step.

        This enables optimization algorithms that perform several evaluations
        of the gradient before taking a step (ie. second order methods like
        LBFGS).

        Additionally, this is where:

        - ``metamer_representation`` is calculated, and thus any modifications
          to the model's forward call (e.g., specifying ``scale`` kwarg for
          coarse-to-fine) should happen.

        - ``loss`` is calculated and ``loss.backward()`` is called.

        Returns
        -------
        loss
            Loss of the current objective function.
        """
        self.optimizer.zero_grad()
        analyze_kwargs = {}
        # if we've reached 'all', we use the full model
        if self.scales[0] != "all":
            analyze_kwargs["scales"] = [self.scales[0]]
            # if 'together', then we also want all the coarser
            # scales
            if self.coarse_to_fine == "together":
                analyze_kwargs["scales"] += self.scales_finished
        # if analyze_kwargs is empty, we can just compare
        # metamer_representation against our cached target_representation
        if analyze_kwargs:
            if self._ctf_target_representation is None:
                target_rep = self.model(self.image, **analyze_kwargs)
                self._ctf_target_representation = target_rep
            else:
                target_rep = self._ctf_target_representation
        else:
            target_rep = None

        loss = self.objective_function(self.metamer, target_rep, **analyze_kwargs)
        loss.backward(retain_graph=False)

        return loss.item()

    def _check_convergence(
        self,
        i: int,
        stop_criterion: float,
        stop_iters_to_check: int,
        ctf_iters_to_check: int,
    ) -> bool:
        r"""
        Check whether the loss has stabilized and whether we've synthesized all scales.

        We check whether:

        - We have been synthesizing for ``stop_iters_to_check`` iterations,
          i.e. ``len(synth.losses) > stop_iters_to_check``.

        - Loss has decreased by less than ``stop_criterion`` over the past
          ``stop_iters_to_check`` iterations.

        - We have finished synthesizing each individual scale, i.e. ``synth.scales[0] ==
          "all"``.

        - We have been synthesizing all scales for more than ``ctf_iters_to_check``
          iterations, i.e. ``i - synth.scales_timing["all"][0]) > ctf_iters_to_check``.

        If all conditions are met, we return ``True``. Else, we return ``False``.

        Parameters
        ----------
        i
            The current iteration (0-indexed).
        stop_criterion
            If the loss over the past ``stop_iters_to_check`` has changed
            less than ``stop_criterion``, we terminate synthesis.
        stop_iters_to_check
            How many iterations back to check in order to see if the
            loss has stopped decreasing (for ``stop_criterion``).
        ctf_iters_to_check
            Minimum number of iterations coarse-to-fine must run at each scale.

        Returns
        -------
        loss_stabilized
            Whether the loss has stabilized and we've synthesized all scales.
        """  # noqa: E501
        loss_conv = loss_convergence(self, stop_criterion, stop_iters_to_check)
        return loss_conv and coarse_to_fine_enough(self, i, ctf_iters_to_check)

    def to(self, *args: Any, **kwargs: Any):
        r"""
        Move and/or cast the parameters and buffers.

        This can be called as

        .. function:: to(device=None, dtype=None, non_blocking=False)

        .. function:: to(dtype, non_blocking=False)

        .. function:: to(tensor, non_blocking=False)

        Its signature is similar to :meth:`torch.Tensor.to`, but only accepts
        floating point desired :attr:`dtype` s. In addition, this method will
        only cast the floating point parameters and buffers to :attr:`dtype`
        (if given). The integral parameters and buffers will be moved
        :attr:`device`, if that is given, but with dtypes unchanged. When
        :attr:`non_blocking` is set, it tries to convert/move asynchronously
        with respect to the host if possible, e.g., moving CPU Tensors with
        pinned memory to CUDA devices.

        See :meth:`torch.nn.Module.to` for examples.

        .. note::
            This method modifies the module in-place.

        Parameters
        ----------
        device : torch.device
            The desired device of the parameters and buffers in this module.
        dtype : torch.dtype
            The desired floating point type of the floating point parameters and
            buffers in this module.
        tensor : torch.Tensor
            Tensor whose dtype and device are the desired dtype and device for
            all parameters and buffers in this module.
        """  # numpydoc ignore=PR01,PR02
        super().to(*args, **kwargs)
        # if synthesize has been called at least once and we have not finished moving
        # through all scales, _ctf_target_representation will be a Tensor which get
        # passed to objective_function at some point. thus, need to make sure it's also
        # updated.
        if self._ctf_target_representation is not None:
            self._ctf_target_representation = self._ctf_target_representation.to(
                *args, **kwargs
            )

    def load(
        self,
        file_path: str,
        map_location: str | None = None,
        tensor_equality_atol: float = 1e-8,
        tensor_equality_rtol: float = 1e-5,
        **pickle_load_args: Any,
    ):
        r"""
        Load all relevant stuff from a .pt file.

        This should be called by an initialized ``Metamer`` object -- we will
        ensure that ``image``, ``target_representation`` (and thus
        ``model``), and ``loss_function`` are all identical.

        Note this operates in place and so doesn't return anything.

        Parameters
        ----------
        file_path
            The path to load the synthesis object from.
        map_location
            Argument to pass to :func:`torch.load` as ``map_location``. If you
            save stuff that was being run on a GPU and are loading onto a
            CPU, you'll need this to make sure everything lines up
            properly. This should be structured like the str you would
            pass to :class:`torch.device`.
        tensor_equality_atol
            Absolute tolerance to use when checking for tensor equality during load,
            passed to :func:`torch.allclose`. It may be necessary to increase if you are
            saving and loading on two machines with torch built by different cuda
            versions. Be careful when changing this! See
            :class:`torch.finfo<torch.torch.finfo>` for more details about floating
            point precision of different data types (especially, ``eps``); if you have
            to increase this by more than 1 or 2 decades, then you are probably not
            dealing with a numerical issue.
        tensor_equality_rtol
            Relative tolerance to use when checking for tensor equality during load,
            passed to :func:`torch.allclose`. It may be necessary to increase if you are
            saving and loading on two machines with torch built by different cuda
            versions. Be careful when changing this! See
            :class:`torch.finfo<torch.torch.finfo>` for more details about floating
            point precision of different data types (especially, ``eps``); if you have
            to increase this by more than 1 or 2 decades, then you are probably not
            dealing with a numerical issue.
        **pickle_load_args
            Any additional kwargs will be added to ``pickle_module.load`` via
            :func:`torch.load`, see that function's docstring for details.

        Raises
        ------
        ValueError
            If :func:`setup` or :func:`synthesize` has been called before this call
            to ``load``.
        ValueError
            If the object saved at ``file_path`` is not a ``MetamerCTF`` object.
        ValueError
            If the saved and loading ``MetamerCTF`` objects have a different value
            for any of :attr:`image`, :attr:`range_penalty_lambda`,
            :attr:`allowed_range`, or :attr:`coarse_to_fine`.
        ValueError
            If the behavior of :attr:`loss_function` or :attr:`model` is different
            between the saved and loading objects.

        Warns
        -----
        UserWarning
            If :func:`setup` will need to be called after load, to finish initializing
            :attr:`optimizer` or :attr:`scheduler`.

        Examples
        --------
        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> model = po.simul.PortillaSimoncelli(img.shape[-2:])
        >>> metamer = po.synth.MetamerCTF(img, model)
        >>> metamer.synthesize(max_iter=5, store_progress=True)
        >>> metamer.save("metamers_ctf.pt")
        >>> metamer_copy = po.synth.MetamerCTF(img, model)
        >>> metamer_copy.load("metamers_ctf.pt")
        """
        super()._load(
            file_path,
            map_location,
            ["_coarse_to_fine"],
            tensor_equality_atol=tensor_equality_atol,
            tensor_equality_rtol=tensor_equality_rtol,
            **pickle_load_args,
        )

    @property
    def coarse_to_fine(self) -> str:
        """How we scales are handled, see :class:`MetamerCTF` for details."""
        # numpydoc ignore=RT01,ES01
        return self._coarse_to_fine

    @property
    def scales(self) -> tuple:
        """Model scales that we've yet to optimize, modified during optimization."""
        # numpydoc ignore=RT01,ES01
        return tuple(self._scales)

    @property
    def scales_loss(self) -> tuple:
        """Scale-specific loss at each iteration."""
        # numpydoc ignore=RT01,ES01
        return tuple(self._scales_loss)

    @property
    def scales_timing(self) -> dict:
        """
        Information about when each scale was started and stopped.

        Keys are the values found in :attr:`scales`, and values are lists specifying
        the iteration where we started and stopped optimizing this scale, which are
        modified during optimization.
        """  # numpydoc ignore=RT01
        return self._scales_timing

    @property
    def scales_finished(self) -> tuple:
        """Model scales that we've finished optimizing, modified during optimization."""
        # numpydoc ignore=RT01,ES01
        return tuple(self._scales_finished)


def plot_loss(
    metamer: Metamer,
    iteration: int | None = None,
    ax: mpl.axes.Axes | None = None,
    **kwargs: Any,
) -> mpl.axes.Axes:
    """
    Plot synthesis loss with log-scaled y axis.

    Plots ``metamer.losses`` over all iterations. Also plots a red dot at
    ``iteration``, to highlight the loss there. If ``iteration=None``, then the
    dot will be at the final iteration.

    Parameters
    ----------
    metamer
        Metamer object whose loss we want to plot.
    iteration
        Which iteration to display. If ``None``,  we show
        the most recent one. Negative values are also allowed.
    ax
        Pre-existing axes for plot. If ``None``, we call
        :func:`matplotlib.pyplot.gca()`.
    **kwargs
        Passed to :func:`matplotlib.pyplot.semilogy`.

    Returns
    -------
    ax
        The matplotlib axes containing the plot.

    Raises
    ------
    IndexError
        If ``iteration`` takes an illegal value.
    """
    # this warning is not relevant for this plotting function
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="loss iteration and iteration for")
        progress = metamer.get_progress(iteration)

    if ax is None:
        ax = plt.gca()
    ax.semilogy(metamer.losses, **kwargs)
    ax.scatter(progress["iteration"], progress["losses"], c="r")
    ax.set(xlabel="Synthesis iteration", ylabel="Loss")
    return ax


def display_metamer(
    metamer: Metamer,
    batch_idx: int = 0,
    channel_idx: int | None = None,
    zoom: float | None = None,
    iteration: int | None = None,
    ax: mpl.axes.Axes | None = None,
    **kwargs: Any,
) -> mpl.axes.Axes:
    """
    Display metamer.

    We use :func:`~plenoptic.tools.display.imshow` to display the metamer and attempt to
    automatically find the most reasonable zoom value. You can override this
    value using the zoom arg, but remember that :func:`~plenoptic.tools.display.imshow`
    is opinionated about the size of the resulting image and will throw an
    exception if the axis created is not big enough for the selected zoom.

    Parameters
    ----------
    metamer
        Metamer object whose synthesized metamer we want to display.
    batch_idx
        Which index to take from the batch dimension.
    channel_idx
        Which index to take from the channel dimension. If ``None``, we assume
        image is RGB(A) and show all channels.
    zoom
        How much to zoom in / enlarge the metamer, the ratio of display pixels
        to image pixels. If ``None``, we attempt to find the best
        value ourselves.
    iteration
        Which iteration to display. If ``None``, we show the most recent one.
        Negative values are also allowed. If ``iteration!=None`` and
        ``metamer.store_progress>1`` (that is, the metamer was not cached on every
        iteration), then we show the cached metamer from the nearest iteration.
    ax
        Pre-existing axes for plot. If ``None``, we call :func:`matplotlib.pyplot.gca`.
    **kwargs
        Passed to :func:`~plenoptic.tools.display.imshow`.

    Returns
    -------
    ax
        The matplotlib axes containing the plot.

    Raises
    ------
    ValueError
        If ``batch_idx`` is not an int.
    IndexError
        If ``iteration`` takes an illegal value.

    Warns
    -----
    UserWarning
        If the iteration for the displayed metamer is not the same as the argument
        ``iteration`` (because e.g., you set ``iteration=3`` but
        ``metamer.store_progress=2``).
    """
    progress = metamer.get_progress(iteration)
    try:
        image = progress["saved_metamer"]
        iter = progress["store_progress_iteration"]
    except KeyError:
        if iteration is not None:
            raise IndexError(
                "When metamer.store_progress=False, iteration must be None!"
            )
        image = metamer.metamer
        # losses will always have one extra value, the current loss.
        iter = len(metamer.losses) - 1

    if not isinstance(batch_idx, int):
        raise ValueError("batch_idx must be an integer!")
    # we're only plotting one image here, so if the user wants multiple
    # channels, they must be RGB
    as_rgb = bool(channel_idx is None and image.shape[1] > 1)
    if ax is None:
        ax = plt.gca()
    display.imshow(
        image,
        ax=ax,
        title=f"Metamer [iteration={iter}]",
        zoom=zoom,
        batch_idx=batch_idx,
        channel_idx=channel_idx,
        as_rgb=as_rgb,
        **kwargs,
    )
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    return ax


def _representation_error(
    metamer: Metamer,
    iteration: int | None = None,
    iteration_selection: Literal["floor", "ceiling", "round"] = "round",
    **kwargs: Any,
) -> Tensor:
    r"""
    Get the representation error.

    This is ``metamer.model(metamer) - target_representation)``. If
    ``iteration`` is not ``None``, we use
    ``metamer.model(saved_metamer[iteration])`` instead.

    Parameters
    ----------
    metamer
        Metamer object whose representation error we want to compute.
    iteration
        Which iteration to display. If ``None``, we show the most recent one.
        Negative values are also allowed. If ``iteration!=None`` and
        ``metamer.store_progress>1`` (that is, the metamer was not cached on every
        iteration), then we show the cached metamer from the nearest iteration.
    iteration_selection

        How to select the relevant iteration from :attr:`saved_metamer`
        when the request iteration wasn't stored.

        When synthesis was run with ``store_progress=n`` (where ``n>1``),
        metamers are only saved every ``n`` iterations. If you request an
        iteration where a metamer wasn't saved, this determines which available
        iteration is used instead:

        * ``"floor"``: use the closest saved iteration **before** the
          requested one.

        * ``"ceiling"``: use the closest saved iteration **after** the
          requested one.

        * ``"round"``: use the closest saved iteration.

    **kwargs
        Passed to ``metamer.model.forward``.

    Returns
    -------
    representation_error
        The representation error at the specified iteration, for displaying.

    Raises
    ------
    IndexError
        If ``iteration`` takes an illegal value.

    Warns
    -----
    UserWarning
        If the iteration for the used metamer is not the same as the argument
        ``iteration`` (because e.g., you set ``iteration=3`` but
        ``metamer.store_progress=2``).
    """
    if iteration is not None:
        progress = metamer.get_progress(iteration)
        image = progress["saved_metamer"].to(metamer.target_representation.device)
        metamer_rep = metamer.model(image, **kwargs)
    else:
        metamer_rep = metamer.model(metamer.metamer, **kwargs)
    return metamer_rep - metamer.target_representation


def plot_representation_error(
    metamer: Metamer,
    batch_idx: int = 0,
    iteration: int | None = None,
    ylim: tuple[float, float] | None | Literal[False] = None,
    ax: mpl.axes.Axes | None = None,
    as_rgb: bool = False,
    **kwargs: Any,
) -> list[mpl.axes.Axes]:
    r"""
    Plot distance ratio showing how close we are to convergence.

    We plot ``_representation_error(metamer, iteration)``. For more details, see
    :func:`plenoptic.tools.display.plot_representation`.

    Parameters
    ----------
    metamer
        Metamer object whose synthesized metamer we want to display.
    batch_idx
        Which index to take from the batch dimension.
    iteration
        Which iteration to display. If ``None``, we show the most recent one.
        Negative values are also allowed. If ``iteration!=None`` and
        ``metamer.store_progress>1`` (that is, the metamer was not cached on every
        iteration), then we show the cached metamer from the nearest iteration.
    ylim
        If ``ylim`` is ``None``, we sets the axes' y-limits to be ``(-y_max,
        y_max)``, where ``y_max=np.abs(data).max()``. If it's ``False``, we do
        nothing. If a tuple, we use that range.
    ax
        Pre-existing axes for plot. If ``None``, we call :func:`matplotlib.pyplot.gca`.
    as_rgb
        The representation can be image-like with multiple channels, and we
        have no way to determine whether it should be represented as an RGB
        image or not, so the user must set this flag to tell us. It will be
        ignored if the response doesn't look image-like or if the model has its
        own ``plot_representation_error()`` method. Else, it will be passed to
        :func:`~plenoptic.tools.display.imshow`, see that methods docstring for details.
    **kwargs
        Passed to ``metamer.model.forward``.

    Returns
    -------
    axes :
        List of created axes.

    Raises
    ------
    IndexError
        If ``iteration`` takes an illegal value.

    Warns
    -----
    UserWarning
        If the iteration for the metamer used to compute the error is not the same as
        the argument ``iteration`` (because e.g., you set ``iteration=3`` but
        ``metamer.store_progress=2``).
    """
    representation_error = _representation_error(
        metamer=metamer, iteration=iteration, **kwargs
    )
    if ax is None:
        ax = plt.gca()
    return display.plot_representation(
        metamer.model,
        representation_error,
        ax,
        title="Representation error",
        ylim=ylim,
        batch_idx=batch_idx,
        as_rgb=as_rgb,
    )


def plot_pixel_values(
    metamer: Metamer,
    batch_idx: int = 0,
    channel_idx: int | None = None,
    iteration: int | None = None,
    ylim: tuple[float, float] | Literal[False] = False,
    ax: mpl.axes.Axes | None = None,
    **kwargs: Any,
) -> mpl.axes.Axes:
    r"""
    Plot histogram of pixel values of target image and its metamer.

    As a way to check the distributions of pixel intensities and see
    if there's any values outside the allowed range

    Parameters
    ----------
    metamer
        Metamer object with the images whose pixel values we want to compare.
    batch_idx
        Which index to take from the batch dimension.
    channel_idx
        Which index to take from the channel dimension. If ``None``, we use all
        channels (assumed use-case is RGB(A) images).
    iteration
        Which iteration to display. If ``None``, we show the most recent one.
        Negative values are also allowed. If ``iteration!=None`` and
        ``metamer.store_progress>1`` (that is, the metamer was not cached on every
        iteration), then we show the cached metamer from the nearest iteration.
    ylim
        If tuple, the ylimit to set for this axis. If ``False``, we leave
        it untouched.
    ax
        Pre-existing axes for plot. If ``None``, we call
        :func:`matplotlib.pyplot.gca()`.
    **kwargs
        Passed to :func:`matplotlib.pyplot.hist`.

    Returns
    -------
    ax
        Created axes.

    Raises
    ------
    IndexError
        If ``iteration`` takes an illegal value.

    Warns
    -----
    UserWarning
        If the iteration used for ``saved_metamer`` is not the same as the argument
        ``iteration`` (because e.g., you set ``iteration=3`` but
        ``metamer.store_progress=2``).
    """

    def _freedman_diaconis_bins(a: np.ndarray) -> int:
        """
        Calculate number of hist bins using Freedman-Diaconis rule.

        Copied from seaborn.

        Parameters
        ----------
        a
            The array to histogram.

        Returns
        -------
        n_bins
            Number of bins to use for histogram.
        """
        # From https://stats.stackexchange.com/questions/798/
        a = np.asarray(a)
        iqr = np.diff(np.percentile(a, [0.25, 0.75]))[0]
        if len(a) < 2:
            return 1
        h = 2 * iqr / (len(a) ** (1 / 3))
        # fall back to sqrt(a) bins if iqr is 0
        if h == 0:
            return int(np.sqrt(a.size))
        else:
            return int(np.ceil((a.max() - a.min()) / h))

    kwargs.setdefault("alpha", 0.4)
    progress = metamer.get_progress(iteration)
    try:
        met = progress["saved_metamer"]
        iter = progress["store_progress_iteration"]
    except KeyError:
        if iteration is not None:
            raise IndexError(
                "When metamer.store_progress=False, iteration must be None!"
            )
        met = metamer.metamer
        # losses will always have one extra value, the current loss.
        iter = len(metamer.losses) - 1
    image = metamer.image[batch_idx]
    if channel_idx is not None:
        image = image[channel_idx]
        met = met[channel_idx]
    image = data.to_numpy(image).flatten()
    met = data.to_numpy(met).flatten()

    if ax is None:
        ax = plt.gca()
    ax.hist(
        met,
        bins=min(_freedman_diaconis_bins(image), 50),
        label=f"Metamer [iteration={iter}]",
        **kwargs,
    )
    ax.hist(
        image,
        bins=min(_freedman_diaconis_bins(image), 50),
        label="Target image",
        **kwargs,
    )
    ax.legend()
    if ylim:
        ax.set_ylim(ylim)
    ax.set_title("Histogram of pixel values")
    return ax


def _check_included_plots(to_check: list[str] | dict[str, float], to_check_name: str):
    """
    Check whether the user wanted us to create plots that we can't.

    Helper function for :func:`plot_synthesis_status` and :func:`animate`.

    Raises a ``ValueError`` if ``to_check`` contains any values that are not allowed.

    Parameters
    ----------
    to_check
        The variable to check. We ensure that it doesn't contain any extra (not
        allowed) values. If a list, we check its contents. If a dict, we check
        its keys.
    to_check_name
        Name of the ``to_check`` variable, used in the error message.

    Raises
    ------
    ValueError
        If ``to_check`` takes an illegal value.
    """
    allowed_vals = [
        "display_metamer",
        "plot_loss",
        "plot_representation_error",
        "plot_pixel_values",
        "misc",
    ]
    try:
        vals = to_check.keys()
    except AttributeError:
        vals = to_check
    not_allowed = [v for v in vals if v not in allowed_vals]
    if not_allowed:
        raise ValueError(
            f"{to_check_name} contained value(s) {not_allowed}! "
            f"Only {allowed_vals} are permissible!"
        )


def _setup_synthesis_fig(
    fig: mpl.figure.Figure | None = None,
    axes_idx: dict[str, int] = {},
    figsize: tuple[float, float] | None = None,
    included_plots: list[str] = [
        "display_metamer",
        "plot_loss",
        "plot_representation_error",
    ],
    display_metamer_width: float = 1,
    plot_loss_width: float = 1,
    plot_representation_error_width: float = 1,
    plot_pixel_values_width: float = 1,
) -> tuple[mpl.figure.Figure, list[mpl.axes.Axes], dict[str, int]]:
    """
    Set up figure for :func:`plot_synthesis_status`.

    Creates figure with enough axes for the all the plots you want. Will
    also create index in ``axes_idx`` for them if you haven't done so already.

    If ``fig=None``, all axes will be on the same row and have the same width.
    If you want them to be on different rows, will need to initialize ``fig``
    yourself and pass that in. For changing width, change the corresponding
    ``*_width`` arg, which gives width relative to other axes. So if you want
    the axis for the ``representation_error`` plot to be twice as wide as the
    others, set ``representation_error_width=2``.

    Parameters
    ----------
    fig
        The figure to plot on or ``None``. If ``None``, we create a new figure.
    axes_idx
        Dictionary specifying which axes contains which type of plot, allows for more
        fine-grained control of the resulting figure. Probably only helpful if fig is
        also defined. Possible keys: ``"loss"``, ``"representation_error"``,
        ``"pixel_values"``, ``"misc"``. Values should all be ints. If you tell this
        function to create a plot that doesn't have a corresponding key, we find the
        lowest int that is not already in the dict, so if you have
        axes that you want unchanged, place their idx in ``"misc"``.
    figsize
        The size of the figure to create. It may take a little bit of
        playing around to find a reasonable value. If ``None``, we attempt to
        make our best guess, aiming to have relative width=1 correspond to 5.
    included_plots
        Which plots to include. Must be some subset of ``'display_metamer',
        'plot_loss', 'plot_representation_error', 'plot_pixel_values'``.
    display_metamer_width
        Relative width of the axis for the synthesized metamer.
    plot_loss_width
        Relative width of the axis for loss plot.
    plot_representation_error_width
        Relative width of the axis for representation error plot.
    plot_pixel_values_width
        Relative width of the axis for image pixel intensities histograms.

    Returns
    -------
    fig
        The figure to plot on.
    axes
        List or array of axes contained in fig.
    axes_idx
        Dictionary identifying the idx for each plot type.
    """
    n_subplots = 0
    axes_idx = axes_idx.copy()
    width_ratios = []
    if "display_metamer" in included_plots:
        n_subplots += 1
        width_ratios.append(display_metamer_width)
        if "display_metamer" not in axes_idx:
            axes_idx["display_metamer"] = data._find_min_int(axes_idx.values())
    if "plot_loss" in included_plots:
        n_subplots += 1
        width_ratios.append(plot_loss_width)
        if "plot_loss" not in axes_idx:
            axes_idx["plot_loss"] = data._find_min_int(axes_idx.values())
    if "plot_representation_error" in included_plots:
        n_subplots += 1
        width_ratios.append(plot_representation_error_width)
        if "plot_representation_error" not in axes_idx:
            axes_idx["plot_representation_error"] = data._find_min_int(
                axes_idx.values()
            )
    if "plot_pixel_values" in included_plots:
        n_subplots += 1
        width_ratios.append(plot_pixel_values_width)
        if "plot_pixel_values" not in axes_idx:
            axes_idx["plot_pixel_values"] = data._find_min_int(axes_idx.values())
    if fig is None:
        width_ratios = np.array(width_ratios)
        if figsize is None:
            # we want (5, 5) for each subplot, with a bit of room between
            # each subplot
            figsize = ((width_ratios * 5).sum() + width_ratios.sum() - 1, 5)
        width_ratios = width_ratios / width_ratios.sum()
        fig, axes = plt.subplots(
            1,
            n_subplots,
            figsize=figsize,
            gridspec_kw={"width_ratios": width_ratios},
        )
        if n_subplots == 1:
            axes = [axes]
    else:
        axes = fig.axes
    # make sure misc contains all the empty axes
    misc_axes = axes_idx.get("misc", [])
    if not hasattr(misc_axes, "__iter__"):
        misc_axes = [misc_axes]
    all_axes = []
    for i in axes_idx.values():
        # so if it's a list of ints
        if hasattr(i, "__iter__"):
            all_axes.extend(i)
        else:
            all_axes.append(i)
    misc_axes += [i for i, _ in enumerate(fig.axes) if i not in all_axes]
    axes_idx["misc"] = misc_axes
    return fig, axes, axes_idx


def plot_synthesis_status(
    metamer: Metamer,
    batch_idx: int = 0,
    channel_idx: int | None = None,
    iteration: int | None = None,
    ylim: tuple[float, float] | None | Literal[False] = None,
    vrange: tuple[float, float] | str = "indep1",
    zoom: float | None = None,
    plot_representation_error_as_rgb: bool = False,
    fig: mpl.figure.Figure | None = None,
    axes_idx: dict[str, int] = {},
    figsize: tuple[float, float] | None = None,
    included_plots: list[str] = [
        "display_metamer",
        "plot_loss",
        "plot_representation_error",
    ],
    width_ratios: dict[str, float] = {},
) -> tuple[mpl.figure.Figure, dict[str, int]]:
    r"""
    Make a plot showing synthesis status.

    We create several subplots to analyze this. The plots to include are
    specified by including their name in the ``included_plots`` list. All plots
    can be created separately using the method with the same name.

    Parameters
    ----------
    metamer
        Metamer object whose status we want to plot.
    batch_idx
        Which index to take from the batch dimension.
    channel_idx
        Which index to take from the channel dimension. If ``None``, we use all
        channels (assumed use-case is RGB(A) image).
    iteration
        Which iteration to display. If ``None``, we show the most recent one.
        Negative values are also allowed. If ``iteration!=None`` and
        ``metamer.store_progress>1`` (that is, the metamer was not cached on every
        iteration), then we show the cached metamer from the nearest iteration.
    ylim
        The ylimit to use for the representation_error plot. We pass
        this value directly to ``plot_representation_error``.
    vrange
        The vrange option to pass to :func:`display_metamer()`. See
        docstring of :func:`~plenoptic.tools.display.imshow` for possible values.
    zoom
        How much to zoom in / enlarge the metamer, the ratio
        of display pixels to image pixels. If ``None``, we
        attempt to find the best value ourselves.
    plot_representation_error_as_rgb
        The representation can be image-like with multiple channels, and we
        have no way to determine whether it should be represented as an RGB
        image or not, so the user must set this flag to tell us. It will be
        ignored if the response doesn't look image-like or if the
        model has its own plot_representation_error() method. Else, it will
        be passed to :func:`~plenoptic.tools.display.imshow`, see that method's
        docstring for details.
    fig
        If ``None``, we create a new figure. otherwise we assume this is
        an empty figure that has the appropriate size and number of
        subplots.
    axes_idx
        Dictionary specifying which axes contains which type of plot, allows
        for more fine-grained control of the resulting figure. Probably only
        helpful if fig is also defined. Possible keys: ``'display_metamer',
        'plot_loss', 'plot_representation_error', 'plot_pixel_values',
        'misc'``. Values should all be ints. If you tell this function to
        create a plot that doesn't have a corresponding key, we find the lowest
        int that is not already in the dict, so if you have axes that you want
        unchanged, place their idx in ``'misc'``.
    figsize
        The size of the figure to create. It may take a little bit of
        playing around to find a reasonable value. If ``None``, we attempt to
        make our best guess, aiming to have each axis be of size ``(5, 5)``.
    included_plots
        Which plots to include. Must be some subset of ``'display_metamer',
        'plot_loss', 'plot_representation_error', 'plot_pixel_values'``.
    width_ratios
        If ``width_ratios`` is an empty dictionary, ``plot_loss`` will have
        double the width of the other plots. To change that, specify their
        relative widths using the keys: ['display_mad_image', 'plot_loss',
        'plot_pixel_values'] and floats specifying their relative width.

    Returns
    -------
    fig
        The figure containing this plot.
    axes_idx
        Dictionary giving index of each plot.

    Raises
    ------
    ValueError
        If ``metamer.metamer`` object is not 3d or 4d.
    ValueError
        If the ``iteration is not None`` and the given ``metamer`` object was run
        with ``store_progress=False``.

    Warns
    -----
    UserWarning
        If the iteration used for ``saved_metamer`` is not the same as the argument
        ``iteration`` (because e.g., you set ``iteration=3`` but
        ``metamer.store_progress=2``).
    """
    if iteration is not None and not metamer.store_progress:
        raise ValueError(
            "synthesis() was run with store_progress=False, "
            "cannot specify which iteration to plot (only"
            " last one, with iteration=None)"
        )
    if metamer.metamer.ndim not in [3, 4]:
        raise ValueError(
            "plot_synthesis_status() expects 3 or 4d data;"
            "unexpected behavior will result otherwise!"
        )
    _check_included_plots(included_plots, "included_plots")
    _check_included_plots(width_ratios, "width_ratios")
    _check_included_plots(axes_idx, "axes_idx")
    width_ratios = {f"{k}_width": v for k, v in width_ratios.items()}
    fig, axes, axes_idx = _setup_synthesis_fig(
        fig, axes_idx, figsize, included_plots, **width_ratios
    )

    def check_iterables(i: int, vals: list | tuple) -> bool:
        """
        Determine whether i is in vals.

        Works with an iterable of iterables and iterable of non-iterables.

        Parameters
        ----------
        i
            The value we're looking for.
        vals
            The iterable it might be in.

        Returns
        -------
        contained
            Whether i is in vals.
        """
        for j in vals:
            try:
                # then it's an iterable
                if i in j:
                    return True
            except TypeError:
                # then it's not an iterable
                if i == j:
                    return True

    if "display_metamer" in included_plots:
        display_metamer(
            metamer,
            batch_idx=batch_idx,
            channel_idx=channel_idx,
            iteration=iteration,
            ax=axes[axes_idx["display_metamer"]],
            zoom=zoom,
            vrange=vrange,
        )
    if "plot_loss" in included_plots:
        plot_loss(metamer, iteration=iteration, ax=axes[axes_idx["plot_loss"]])
    if "plot_representation_error" in included_plots:
        plot_representation_error(
            metamer,
            batch_idx=batch_idx,
            iteration=iteration,
            ax=axes[axes_idx["plot_representation_error"]],
            ylim=ylim,
            as_rgb=plot_representation_error_as_rgb,
        )
        # this can add a bunch of axes, so this will try and figure
        # them out
        new_axes = [
            i
            for i, _ in enumerate(fig.axes)
            if not check_iterables(i, axes_idx.values())
        ] + [axes_idx["plot_representation_error"]]
        axes_idx["plot_representation_error"] = new_axes
    if "plot_pixel_values" in included_plots:
        plot_pixel_values(
            metamer,
            batch_idx=batch_idx,
            channel_idx=channel_idx,
            iteration=iteration,
            ax=axes[axes_idx["plot_pixel_values"]],
        )
    return fig, axes_idx


def animate(
    metamer: Metamer,
    framerate: int = 10,
    batch_idx: int = 0,
    channel_idx: int | None = None,
    ylim: str | None | tuple[float, float] | Literal[False] = None,
    vrange: tuple[float, float] | str = (0, 1),
    zoom: float | None = None,
    plot_representation_error_as_rgb: bool = False,
    fig: mpl.figure.Figure | None = None,
    axes_idx: dict[str, int] = {},
    figsize: tuple[float, float] | None = None,
    included_plots: list[str] = [
        "display_metamer",
        "plot_loss",
        "plot_representation_error",
    ],
    width_ratios: dict[str, float] = {},
) -> mpl.animation.FuncAnimation:
    r"""
    Animate synthesis progress.

    This is essentially the figure produced by
    ``metamer.plot_synthesis_status`` animated over time, for each stored
    iteration.

    This functions returns a matplotlib FuncAnimation object. See our
    documentation (e.g., :ref:`quickstart-nb`) for examples on how to view it in
    a Jupyter notebook. In order to save, use ``anim.save(filename)``. In either
    case, this can take a while and you'll need the appropriate writer installed
    and on your path, e.g., ffmpeg, imagemagick, etc). See
    :doc:`matplotlib documentation <matplotlib:api/animation_api>` for more details.

    Parameters
    ----------
    metamer
        Metamer object whose synthesis we want to animate.
    framerate
        How many frames a second to display.
    batch_idx
        Which index to take from the batch dimension.
    channel_idx
        Which index to take from the channel dimension. If ``None``, we use all
        channels (assumed use-case is RGB(A) image).
    ylim
        The y-limits of the representation_error plot:

        * If a tuple, then this is the ylim of all plots

        * If ``None``, then all plots have the same limits, all
          symmetric about 0 with a limit of
          ``np.abs(representation_error).max()`` (for the initial
          representation_error).

        * If ``False``, don't modify limits.

        * If a string, must be ``"rescale"`` or of the form ``"rescaleN"``,
          where N can be any integer. If ``"rescaleN"``, we rescale the
          limits every N frames (we rescale as if ``ylim=None``). If
          ``"rescale"``, then we do this 10 times over the course of the
          animation.

    vrange
        The vrange option to pass to :func:`display_metamer()`. See
        docstring of :func:`~plenoptic.tools.display.imshow` for possible values.
    zoom
        How much to zoom in / enlarge the metamer, the ratio
        of display pixels to image pixels. If ``None``, we
        attempt to find the best value ourselves.
    plot_representation_error_as_rgb
        The representation can be image-like with multiple channels, and we
        have no way to determine whether it should be represented as an RGB
        image or not, so the user must set this flag to tell us. It will be
        ignored if the representation doesn't look image-like or if the
        model has its own ``plot_representation_error()`` method. Else, it will
        be passed to :func:`~plenoptic.tools.display.imshow`, see that method's
        docstring for details.
    fig
        If ``None``, create the figure from scratch. Else, should be an empty
        figure with enough axes (the expected use here is have same-size
        movies with different plots).
    axes_idx
        Dictionary specifying which axes contains which type of plot, allows
        for more fine-grained control of the resulting figure. Probably only
        helpful if fig is also defined. Possible keys: ``'display_metamer',
        'plot_loss', 'plot_representation_error', 'plot_pixel_values',
        'misc'``. Values should all be ints. If you tell this function to
        create a plot that doesn't have a corresponding key, we find the lowest
        int that is not already in the dict, so if you have axes that you want
        unchanged, place their idx in ``'misc'``.
    figsize
        The size of the figure to create. It may take a little bit of
        playing around to find a reasonable value. If ``None``, we attempt to
        make our best guess, aiming to have each axis be of size ``(5, 5)``.
    included_plots :
        Which plots to include. Must be some subset of ``'display_metamer',
        'plot_loss', 'plot_representation_error', 'plot_pixel_values'``.
    width_ratios
        If ``width_ratios`` is an empty dictionary, ``plot_loss`` will have
        double the width of the other plots. To change that, specify their
        relative widths using the keys: ['display_mad_image', 'plot_loss',
        'plot_pixel_values'] and floats specifying their relative width.

    Returns
    -------
    anim
        The animation object. In order to view, must convert to HTML
        or save.

    Raises
    ------
    ValueError
        If the given ``metamer`` object was run with ``store_progress=False``.
    ValueError
        If ``metamer.metamer`` object is not 3d or 4d.
    ValueError
        If we do not know how to interpret the value of ``ylim``.

    Notes
    -----
    Unless specified, we use the ffmpeg backend, which requires that you have
    ffmpeg installed and on your path (https://ffmpeg.org/download.html). To use
    a different, use the matplotlib rcParams:
    ``matplotlib.rcParams['animation.writer'] = writer``, see `matplotlib
    documentation
    <https://matplotlib.org/stable/api/animation_api.html#writer-classes>`_ for
    more details.
    """
    if not metamer.store_progress:
        raise ValueError(
            "synthesize() was run with store_progress=False, cannot animate!"
        )
    if metamer.metamer.ndim not in [3, 4]:
        raise ValueError(
            "animate() expects 3 or 4d data; unexpected behavior will result otherwise!"
        )
    _check_included_plots(included_plots, "included_plots")
    _check_included_plots(width_ratios, "width_ratios")
    _check_included_plots(axes_idx, "axes_idx")
    if metamer.target_representation.ndimension() == 4:
        # we have to do this here so that we set the
        # ylim_rescale_interval such that we never rescale ylim
        # (rescaling ylim messes up an image axis)
        ylim = False
    try:
        if ylim.startswith("rescale"):
            try:
                ylim_rescale_interval = int(ylim.replace("rescale", ""))
            except ValueError:
                # then there's nothing we can convert to an int there
                ylim_rescale_interval = int((metamer.saved_metamer.shape[0] - 1) // 10)
                if ylim_rescale_interval == 0:
                    ylim_rescale_interval = int(metamer.saved_metamer.shape[0] - 1)
            ylim = None
        else:
            raise ValueError(f"Don't know how to handle ylim {ylim}!")
    except AttributeError:
        # this way we'll never rescale
        ylim_rescale_interval = len(metamer.saved_metamer) + 1
    # we run plot_synthesis_status to initialize the figure if either fig is
    # None or if there are no titles on any axes, which we assume means that
    # it's an empty figure
    if fig is None or not any([ax.get_title() for ax in fig.axes]):
        fig, axes_idx = plot_synthesis_status(
            metamer=metamer,
            batch_idx=batch_idx,
            channel_idx=channel_idx,
            iteration=0,
            figsize=figsize,
            ylim=ylim,
            vrange=vrange,
            zoom=zoom,
            fig=fig,
            axes_idx=axes_idx,
            included_plots=included_plots,
            plot_representation_error_as_rgb=plot_representation_error_as_rgb,
            width_ratios=width_ratios,
        )
    # grab the artist for the second plot (we don't need to do this for the
    # metamer or representation plot, because we use the update_plot
    # function for that)
    if "plot_loss" in included_plots:
        scat = fig.axes[axes_idx["plot_loss"]].collections[0]
    # can have multiple plots
    if "plot_representation_error" in included_plots:
        try:
            rep_error_axes = [
                fig.axes[i] for i in axes_idx["plot_representation_error"]
            ]
        except TypeError:
            # in this case, axes_idx['plot_representation_error'] is not iterable and
            # so is a single value
            rep_error_axes = [fig.axes[axes_idx["plot_representation_error"]]]
    else:
        rep_error_axes = []
    if "display_metamer" in included_plots:
        fig.axes[axes_idx["display_metamer"]].set_title("Metamer")

    if metamer.target_representation.ndimension() == 4:
        if "plot_representation_error" in included_plots:
            warnings.warn(
                "Looks like representation is image-like, haven't fully"
                " thought out how to best handle rescaling color ranges yet!"
            )
        # replace the bit of the title that specifies the range,
        # since we don't make any promises about that. we have to do
        # this here because we need the figure to have been created
        for ax in rep_error_axes:
            ax.set_title(re.sub(r"\n range: .* \n", "\n\n", ax.get_title()))

    def movie_plot(i: int) -> list[mpl.artist.Artist]:
        """
        Matplotlib function for animation.

        Update plots for frame ``i``.

        Parameters
        ----------
        i
            The frame to plot.

        Returns
        -------
        artists
            The updated matplotlib artists.
        """
        # this warning is not relevant for animate
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="loss iteration and iteration for"
            )
            artists = []
            if "display_metamer" in included_plots:
                artists.extend(
                    display.update_plot(
                        fig.axes[axes_idx["display_metamer"]],
                        data=metamer.saved_metamer[i],
                        batch_idx=batch_idx,
                    )
                )
            if "plot_representation_error" in included_plots:
                rep_error = _representation_error(
                    metamer,
                    iteration=min(i * metamer.store_progress, len(metamer.losses) - 1),
                )

                # we pass rep_error_axes to update, and we've grabbed
                # the right things above
                artists.extend(
                    display.update_plot(
                        rep_error_axes,
                        batch_idx=batch_idx,
                        model=metamer.model,
                        data=rep_error,
                    )
                )
                # again, we know that rep_error_axes contains all the axes
                # with the representation ratio info
                if (
                    (i + 1) % ylim_rescale_interval == 0
                    and metamer.target_representation.ndimension() == 3
                ):
                    display.rescale_ylim(rep_error_axes, rep_error)

            if "plot_pixel_values" in included_plots:
                # this is the dumbest way to do this, but it's simple --
                # clearing the axes can cause problems if the user has, for
                # example, changed the tick locator or formatter. not sure how
                # to handle this best right now
                fig.axes[axes_idx["plot_pixel_values"]].clear()
                plot_pixel_values(
                    metamer,
                    batch_idx=batch_idx,
                    channel_idx=channel_idx,
                    iteration=min(i * metamer.store_progress, len(metamer.losses) - 1),
                    ax=fig.axes[axes_idx["plot_pixel_values"]],
                )
            if "plot_loss" in included_plots:
                # loss always contains values from every iteration, but everything
                # else will be subsampled.
                x_val = metamer._convert_iteration(i, False)
                scat.set_offsets((x_val, metamer.losses[x_val]))
                artists.append(scat)
            # as long as blitting is True, need to return a sequence of artists
            return artists

    # don't need an init_func, since we handle initialization ourselves
    anim = mpl.animation.FuncAnimation(
        fig,
        movie_plot,
        frames=len(metamer.saved_metamer),
        blit=True,
        interval=1000.0 / framerate,
        repeat=False,
    )
    plt.close(fig)
    return anim



"""Class versions of perceptual metric functions."""  # numpydoc ignore=ES01

import torch
from deprecated.sphinx import deprecated

from .perceptual_distance import normalized_laplacian_pyramid


@deprecated(
    "NLP will be removed soon, use perceptual_distance.normalized_laplacian_pyramid directly",  # noqa: E501
    "1.2.0",
)
class NLP(torch.nn.Module):
    r"""
    Simple class for implementing normalized laplacian pyramid.

    This class just calls
    :func:`~plenoptic.metric.perceptual_distance.normalized_laplacian_pyramid`
    on the image and returns a 3d tensor with the flattened activations.

    NOTE: synthesis using this class will not be the exact same as
    synthesis using the :func:`~plenoptic.metric.perceptual_distance.nlpd` function,
    because the ``nlpd`` function uses the root-mean square of the
    L2 distance (i.e., ``torch.sqrt(torch.mean(x-y)**2))`` as the distance metric
    between representations.

    Model parameters are those used in [1]_, copied from the matlab code used in the
    paper, found at [2]_.

    References
    ----------
    .. [1] Laparra, V., Ballé, J., Berardino, A. and Simoncelli, E.P., 2016. Perceptual
        image quality assessment using a normalized Laplacian pyramid. Electronic
        Imaging, 2016(16), pp.1-6.
    .. [2] `matlab code <https://www.cns.nyu.edu/~lcv/NLPyr/NLP_dist.m>`_
    """

    def __init__(self):
        super().__init__()

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """
        Compute flattened NLP activations.

        WARNING: For now this only supports images with batch and
        channel size 1

        Parameters
        ----------
        image
            Image to pass to normalized_laplacian_pyramid.

        Returns
        -------
        representation
            3d tensor with flattened NLP activations.

        Raises
        ------
        Exception
            If ``image`` has more than one batch or channel.
        """
        if image.shape[0] > 1 or image.shape[1] > 1:
            raise Exception("For now, this only supports batch and channel size 1")
        activations = normalized_laplacian_pyramid(image)
        # activations is a list of tensors, each at a different scale
        # (down-sampled by factors of 2). To combine these into one
        # vector, we need to flatten each of them and then unsqueeze so
        # it is 3d

        return torch.cat([i.flatten() for i in activations]).unsqueeze(0).unsqueeze(0)



"""Simple models to use with plenoptic's synthesis methods."""

# ignore F401 (unused import) and F403 (from module import *)
# ruff: noqa: F401, F403
from .frontend import *
from .naive import *
from .portilla_simoncelli import PortillaSimoncelli



"""
Portilla-Simoncelli texture statistics.

The Portilla-Simoncelli (PS) texture statistics are a set of image
statistics, first described in [1]_, that are proposed as a sufficient set
of measurements for describing visual textures. That is, if two texture
images have the same values for all PS texture stats, humans should
consider them as members of the same family of textures.

References
----------
.. [1] J Portilla and E P Simoncelli. A Parametric Texture Model based on
   Joint Statistics of Complex Wavelet Coefficients. Int'l Journal of
   Computer Vision. 40(1):49-71, October, 2000.
   https://www.cns.nyu.edu/~eero/ABSTRACTS/portilla99-abstract.html
   https://www.cns.nyu.edu/~lcv/texture/
"""

from collections import OrderedDict
from typing import Literal

import einops
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.fft
import torch.nn as nn
from torch import Tensor

from ...tools import signal, stats
from ...tools.data import to_numpy
from ...tools.display import clean_stem_plot, clean_up_axes, update_stem
from ...tools.validate import validate_input
from ..canonical_computations.steerable_pyramid_freq import (
    SCALES_TYPE as PYR_SCALES_TYPE,
)
from ..canonical_computations.steerable_pyramid_freq import (
    SteerablePyramidFreq,
)

SCALES_TYPE = Literal["pixel_statistics"] | PYR_SCALES_TYPE


class PortillaSimoncelli(nn.Module):
    r"""
    Portila-Simoncelli texture statistics.

    The Portilla-Simoncelli (PS) texture statistics are a set of image
    statistics, first described in [1]_, that are proposed as a sufficient set
    of measurements for describing visual textures. That is, if two texture
    images have the same values for all PS texture stats, humans should
    consider them as members of the same family of textures.

    The PS stats are computed based on the
    :class:`~plenoptic.simulate.canonical_computations.steerable_pyramid_freq.SteerablePyramidFreq`
    [2]_. They consist of the local auto-correlations, cross-scale
    (within-orientation) correlations, and cross-orientation (within-scale)
    correlations of both the pyramid coefficients and the local energy (as
    computed by those coefficients). Additionally, they include the first four
    global moments (mean, variance, skew, and kurtosis) of the image and
    down-sampled versions of that image. See the paper and notebook for more
    description.

    Parameters
    ----------
    image_shape
        Shape of input image.
    n_scales
        The number of pyramid scales used to measure the statistics.
    n_orientations
        The number of orientations used to measure the statistics.
    spatial_corr_width
        The width of the spatial cross- and auto-correlation statistics.

    Attributes
    ----------
    scales: list
        The names of the unique scales of coefficients in the pyramid, used for
        coarse-to-fine metamer synthesis.

    Raises
    ------
    ValueError
        If the height or width of ``image`` cannot be divided by 2 ``n_scales``
        times. This is necessary because of how the model handles multiscale
        representations.

    References
    ----------
    .. [1] J Portilla and E P Simoncelli. A Parametric Texture Model based on
       Joint Statistics of Complex Wavelet Coefficients. Int'l Journal of
       Computer Vision. 40(1):49-71, October, 2000.
       https://www.cns.nyu.edu/~eero/ABSTRACTS/portilla99-abstract.html
       https://www.cns.nyu.edu/~lcv/texture/
    .. [2] E P Simoncelli and W T Freeman, "The Steerable Pyramid: A Flexible
       Architecture for Multi-Scale Derivative Computation," Second Int'l Conf
       on Image Processing, Washington, DC, Oct 1995.
    """

    def __init__(
        self,
        image_shape: tuple[int, int],
        n_scales: int = 4,
        n_orientations: int = 4,
        spatial_corr_width: int = 9,
    ):
        super().__init__()

        self.image_shape = image_shape
        if any([(image_shape[-1] / 2**i) % 2 for i in range(n_scales)]) or any(
            [(image_shape[-2] / 2**i) % 2 for i in range(n_scales)]
        ):
            raise ValueError(
                "Because of how the Portilla-Simoncelli model handles "
                "multiscale representations, it only works with images"
                " whose shape can be divided by 2 `n_scales` times."
            )
        self.spatial_corr_width = spatial_corr_width
        self.n_scales = n_scales
        self.n_orientations = n_orientations
        self._pyr = SteerablePyramidFreq(
            self.image_shape,
            height=self.n_scales,
            order=self.n_orientations - 1,
            is_complex=True,
            tight_frame=False,
        )

        self.scales = (
            ["pixel_statistics", "residual_lowpass"]
            + [ii for ii in range(n_scales - 1, -1, -1)]
            + ["residual_highpass"]
        )

        # Dictionary defining shape of the statistics and which scale they're
        # associated with
        scales_shape_dict = self._create_scales_shape_dict()

        # Dictionary defining necessary statistics, that is, those that are not
        # redundant
        self._necessary_stats_dict = self._create_necessary_stats_dict(
            scales_shape_dict
        )
        # turn this into tensor we can use in forward pass. first into a
        # boolean mask...
        _necessary_stats_mask = einops.pack(
            list(self._necessary_stats_dict.values()), "*"
        )[0]
        # then into a tensor of indices
        _necessary_stats_mask = torch.where(_necessary_stats_mask)[0]
        self.register_buffer("_necessary_stats_mask", _necessary_stats_mask)

        # This array is composed of the following values: 'pixel_statistics',
        # 'residual_lowpass', 'residual_highpass' and integer values from 0 to
        # self.n_scales-1. It is the same size as the representation tensor
        # returned by this object's forward method. It must be a numpy array so
        # we can have a mixture of ints and strs (and so we can use np.isin
        # later)
        self._representation_scales = einops.pack(
            list(scales_shape_dict.values()), "*"
        )[0]
        # just select the scales of the necessary stats.
        self._representation_scales = self._representation_scales[
            self._necessary_stats_mask
        ]
        # This model has no trainable parameters, so it's always in eval mode
        self.eval()

    def _create_scales_shape_dict(self) -> OrderedDict:
        """
        Create dictionary defining scales and shape of each stat.

        This dictionary functions as metadata which is used for two main
        purposes:

        - Scale assignment. In order for optimization to work well, we proceed
          in a "coarse-to-fine" manner. That is, we start optimization by only
          considering the statistics related to the lowest frequencies, and
          gradually add in those related to higher and higher frequencies. This
          is similar to blurring the objective function and then gradually
          adding in finer and finer details. The numbers in this dictionary map
          the computed statistics to their corresponding scales, which we use
          in remove_scales to throw away some stats as needed.

        - Redundant stat identification. As described at the bottom of the
          notebook, the model incidentally computes a whole bunch of redundant
          stats, because auto- and cross-correlation matrices have certain
          symmetries. the _create_necessary_stats_dict method accepts the
          dictionary created here as input and uses the values to get the
          shapes of these and insert True/False as necessary.

        Returns
        -------
        scales_shape_dict
           Dictionary defining shape and associated scales of each computed
           statistic. The keys name each statistic, with dummy arrays as
           values. These arrays have the same shape as the stat (excluding
           batch and channel), with values defining which scale they correspond
           to.
        """
        shape_dict = OrderedDict()
        # There are 6 pixel statistics
        shape_dict["pixel_statistics"] = np.array(6 * ["pixel_statistics"])

        # These are the basic building blocks of the scale assignments for many
        # of the statistics calculated by the PortillaSimoncelli model.
        scales = np.arange(self.n_scales)
        # the cross-scale correlations exclude the coarsest scale
        scales_without_coarsest = np.arange(self.n_scales - 1)
        # the statistics computed on the reconstructed bandpass images have an
        # extra scale corresponding to the lowpass residual
        scales_with_lowpass = np.array(
            scales.tolist() + ["residual_lowpass"], dtype=object
        )

        # now we go through each statistic in order and create a dummy array
        # full of 1s with the same shape as the actual statistic (excluding the
        # batch and channel dimensions, as each stat is computed independently
        # across those dimensions). We then multiply it by one of the scales
        # arrays above to turn those 1s into values describing the
        # corresponding scale.

        auto_corr_mag = np.ones(
            (
                self.spatial_corr_width,
                self.spatial_corr_width,
                self.n_orientations,
                self.n_scales,
            ),
            dtype=int,
        )
        # this rearrange call is turning scales from 1d with shape (n_scales, )
        # to 4d with shape (1, 1, n_scales, 1), so that it matches
        # auto_corr_mag. the following rearrange calls do similar.
        auto_corr_mag *= einops.rearrange(scales, "s -> 1 1 1 s")
        shape_dict["auto_correlation_magnitude"] = auto_corr_mag

        shape_dict["skew_reconstructed"] = scales_with_lowpass

        shape_dict["kurtosis_reconstructed"] = scales_with_lowpass

        auto_corr = np.ones(
            (
                self.spatial_corr_width,
                self.spatial_corr_width,
                self.n_scales + 1,
            ),
            dtype=object,
        )
        auto_corr *= einops.rearrange(scales_with_lowpass, "s -> 1 1 s")
        shape_dict["auto_correlation_reconstructed"] = auto_corr

        shape_dict["std_reconstructed"] = scales_with_lowpass

        cross_orientation_corr_mag = np.ones(
            (self.n_orientations, self.n_orientations, self.n_scales),
            dtype=int,
        )
        cross_orientation_corr_mag *= einops.rearrange(scales, "s -> 1 1 s")
        shape_dict["cross_orientation_correlation_magnitude"] = (
            cross_orientation_corr_mag
        )

        mags_std = np.ones((self.n_orientations, self.n_scales), dtype=int)
        mags_std *= einops.rearrange(scales, "s -> 1 s")
        shape_dict["magnitude_std"] = mags_std

        cross_scale_corr_mag = np.ones(
            (self.n_orientations, self.n_orientations, self.n_scales - 1),
            dtype=int,
        )
        cross_scale_corr_mag *= einops.rearrange(scales_without_coarsest, "s -> 1 1 s")
        shape_dict["cross_scale_correlation_magnitude"] = cross_scale_corr_mag

        cross_scale_corr_real = np.ones(
            (self.n_orientations, 2 * self.n_orientations, self.n_scales - 1),
            dtype=int,
        )
        cross_scale_corr_real *= einops.rearrange(scales_without_coarsest, "s -> 1 1 s")
        shape_dict["cross_scale_correlation_real"] = cross_scale_corr_real

        shape_dict["var_highpass_residual"] = np.array(["residual_highpass"])

        return shape_dict

    def _create_necessary_stats_dict(
        self, scales_shape_dict: OrderedDict
    ) -> OrderedDict:
        """
        Create mask specifying the necessary statistics.

        Some of the statistics computed by the model are redundant, due to
        symmetries. For example, about half of the values in the
        autocorrelation matrices are duplicates. See the Portilla-Simoncelli
        notebook for more details.

        Parameters
        ----------
        scales_shape_dict
            Dictionary defining shape and associated scales of each computed
            statistic.

        Returns
        -------
        necessary_stats_dict
            Dictionary defining which statistics are necessary (i.e., not
            redundant). Will have the same keys as scales_shape_dict, with the
            values being boolean tensors of the same shape as
            scales_shape_dict's corresponding values. True denotes the
            statistics that will be included in the model's output, while False
            denotes the redundant ones we will toss.
        """
        mask_dict = scales_shape_dict.copy()
        # Pre-compute some necessary indices.
        # Lower triangular indices (including diagonal), for auto correlations
        tril_inds = torch.tril_indices(self.spatial_corr_width, self.spatial_corr_width)
        # Get the second half of the diagonal, i.e., everything from the center
        # element on. These are all repeated for the auto correlations. (As
        # these are autocorrelations (rather than auto-covariance) matrices,
        # they've been normalized by the variance and so the center element is
        # always 1, and thus uninformative)
        diag_repeated = torch.arange(
            start=self.spatial_corr_width // 2, end=self.spatial_corr_width
        )
        # Upper triangle indices, including diagonal. These are redundant stats
        # for cross_orientation_correlation_magnitude (because we've normalized
        # this matrix to be true cross-correlations, the diagonals are all 1,
        # like for the auto-correlations)
        triu_inds = torch.triu_indices(self.n_orientations, self.n_orientations)
        for k, v in mask_dict.items():
            if k in [
                "auto_correlation_magnitude",
                "auto_correlation_reconstructed",
            ]:
                # Symmetry M_{i,j} = M_{n-i+1, n-j+1}
                # Start with all False, then place True in necessary stats.
                mask = torch.zeros(v.shape, dtype=torch.bool)
                mask[tril_inds[0], tril_inds[1]] = True
                # if spatial_corr_width is even, then the first row is not
                # redundant with anything either
                if np.mod(self.spatial_corr_width, 2) == 0:
                    mask[0] = True
                mask[diag_repeated, diag_repeated] = False
            elif k == "cross_orientation_correlation_magnitude":
                # Symmetry M_{i,j} = M_{j,i}.
                # Start with all True, then place False in redundant stats.
                mask = torch.ones(v.shape, dtype=torch.bool)
                mask[triu_inds[0], triu_inds[1]] = False
            else:
                # all of the other stats have no redundancies
                mask = torch.ones(v.shape, dtype=torch.bool)
            mask_dict[k] = mask
        return mask_dict

    def forward(self, image: Tensor, scales: list[SCALES_TYPE] | None = None) -> Tensor:
        r"""
        Generate Texture Statistics representation of an image.

        Note that separate batches and channels are analyzed in parallel.

        For any representation that contains info across scales, the scales always run
        from fine to coarse, representing all orientations at a given scale before
        moving on.

        Parameters
        ----------
        image
            A 4d tensor (batch, channel, height, width) containing the image(s) to
            analyze.
        scales
            Which scales to include in the returned representation. If None, we
            include all scales. Otherwise, can contain subset of values present
            in this model's ``scales`` attribute, and the returned tensor will
            then contain the subset corresponding to those scales.

        Returns
        -------
        representation_tensor
            3d tensor of shape (batch, channel, stats) containing the measured
            texture statistics.

        Raises
        ------
        ValueError
            If ``image`` is not 4d or has a dtype other than float or complex.

        Examples
        --------
        >>> import plenoptic as po
        >>> img = po.data.curie()
        >>> portilla_simoncelli_model = po.simul.PortillaSimoncelli(img.shape[2:])
        >>> representation_tensor = portilla_simoncelli_model(img)
        >>> representation_tensor.shape
        torch.Size([1, 1, 1046])
        """
        validate_input(image)

        # pyr_dict is the dictionary of complex-valued tensors returned by the
        # steerable pyramid. pyr_coeffs is a list (length n_scales) of 5d
        # tensors, each of shape (batch, channel, scales, n_orientations,
        # height, width) containing the complex-valued oriented bands, while
        # highpass is a real-valued 4d tensor of shape (batch, channel, height,
        # width). Note that the residual lowpass in pyr_dict has been demeaned.
        # We keep both the dict and list of pyramid coefficients because we
        # need the dictionary for reconstructing the image done later on.
        pyr_dict, pyr_coeffs, highpass, _ = self._compute_pyr_coeffs(image)

        # Now, we create several intermediate representations that we'll use to
        # compute the texture statistics later.

        # First, two intermediate dictionaries: magnitude_pyr_coeffs and
        # real_pyr_coeffs, which contain the demeaned magnitude of the pyramid
        # coefficients and the real part of the pyramid coefficients
        # respectively.
        (
            mag_pyr_coeffs,
            real_pyr_coeffs,
        ) = self._compute_intermediate_representations(pyr_coeffs)

        # Then, the reconstructed lowpass image at each scale. (this is a list
        # of length n_scales+1 containing tensors of shape (batch, channel,
        # height, width))
        reconstructed_images = self._reconstruct_lowpass_at_each_scale(pyr_dict)
        # the reconstructed_images list goes from coarse-to-fine, but we want
        # each of the stats computed from it to go from fine-to-coarse, so we
        # reverse its direction.
        reconstructed_images = reconstructed_images[::-1]

        # Now, start calculating the PS texture stats.

        # Calculate pixel statistics (mean, variance, skew, kurtosis, min,
        # max).
        pixel_stats = self._compute_pixel_stats(image)

        # Compute the central autocorrelation of the coefficient magnitudes. This is a
        # tensor of shape: (batch, channel, spatial_corr_width, spatial_corr_width,
        # n_orientations, n_scales). var_mags is a tensor of shape (batch, channel,
        # n_orientations, n_scales)
        autocorr_mags, mags_var = self._compute_autocorr(mag_pyr_coeffs)
        # mags_var is the variance of the magnitude coefficients at each scale (it's an
        # intermediary of the computation of the auto-correlations). We take the square
        # root to get the standard deviation.
        mags_std = mags_var.sqrt()

        # Compute the central autocorrelation of the reconstructed lowpass
        # images at each scale (and their variances). autocorr_recon is a
        # tensor of shape (batch, channel, spatial_corr_width,
        # spatial_corr_width, n_scales+1), and var_recon is a tensor of shape
        # (batch, channel, n_scales+1)
        autocorr_recon, var_recon = self._compute_autocorr(reconstructed_images)
        # Compute the standard deviation, skew, and kurtosis of each
        # reconstructed lowpass image. std_recon, skew_recon, and
        # kurtosis_recon will all end up as tensors of shape (batch, channel,
        # n_scales+1)
        std_recon = var_recon.sqrt()
        skew_recon, kurtosis_recon = self._compute_skew_kurtosis_recon(
            reconstructed_images, var_recon, pixel_stats[..., 1]
        )

        # Compute the cross-orientation correlations between the magnitude
        # coefficients at each scale. this will be a tensor of shape (batch,
        # channel, n_orientations, n_orientations, n_scales)
        cross_ori_corr_mags = self._compute_cross_correlation(
            mag_pyr_coeffs, mag_pyr_coeffs, mags_var, mags_var
        )

        # If we have more than one scale, compute the cross-scale correlations
        if self.n_scales != 1:
            # First, double the phase the coefficients, so we can correctly
            # compute correlations across scales.
            (
                phase_doubled_mags,
                phase_doubled_sep,
            ) = self._double_phase_pyr_coeffs(pyr_coeffs)
            # Compute the cross-scale correlations between the magnitude
            # coefficients. For each coefficient, we're correlating it with the
            # coefficients at the next-coarsest scale. this will be a tensor of
            # shape (batch, channel, n_orientations, n_orientations,
            # n_scales-1)
            cross_scale_corr_mags = self._compute_cross_correlation(
                mag_pyr_coeffs[:-1], phase_doubled_mags, mags_var[..., :-1]
            )
            # Compute the cross-scale correlations between the real
            # coefficients and the real and imaginary coefficients at the next
            # coarsest scale. this will be a tensor of shape (batch, channel,
            # n_orientations, 2*n_orientations, n_scales-1)
            cross_scale_corr_real = self._compute_cross_correlation(
                real_pyr_coeffs[:-1], phase_doubled_sep
            )

        # Compute the variance of the highpass residual
        var_highpass_residual = highpass.pow(2).mean(dim=(-2, -1))

        # Now, combine all these stats together, first into a list
        all_stats = [
            pixel_stats,
            autocorr_mags,
            skew_recon,
            kurtosis_recon,
            autocorr_recon,
            std_recon,
            cross_ori_corr_mags,
            mags_std,
        ]
        if self.n_scales != 1:
            all_stats += [cross_scale_corr_mags, cross_scale_corr_real]
        all_stats += [var_highpass_residual]
        # And then pack them into a 3d tensor
        representation_tensor, pack_info = einops.pack(all_stats, "b c *")

        # the only time when this is None is during testing, when we make sure
        # that our assumptions are all valid.
        if self._necessary_stats_mask is None:
            # store this so we can unpack this info (only possible when we've
            # discarded no stats)
            self._pack_info = pack_info
        else:
            # Throw away all redundant statistics
            representation_tensor = representation_tensor.index_select(
                -1, self._necessary_stats_mask
            )

        # Return the subset of stats corresponding to the specified scale.
        if scales is not None:
            representation_tensor = self.remove_scales(representation_tensor, scales)

        return representation_tensor

    def remove_scales(
        self, representation_tensor: Tensor, scales_to_keep: list[SCALES_TYPE]
    ) -> Tensor:
        """
        Remove statistics not associated with scales.

        For a given representation_tensor and a list of scales_to_keep, this
        attribute removes all statistics *not* associated with those scales.

        Note that calling this method will always remove statistics.

        Parameters
        ----------
        representation_tensor
            3d tensor containing the measured representation statistics.
        scales_to_keep
            Which scales to include in the returned representation. Can contain
            subset of values present in this model's ``scales`` attribute, and
            the returned tensor will then contain the subset of the full
            representation corresponding to those scales.

        Returns
        -------
        limited_representation_tensor
            Representation tensor with some statistics removed.

        Examples
        --------
        >>> import plenoptic as po
        >>> img = po.data.curie()
        >>> portilla_simoncelli_model = po.simul.PortillaSimoncelli(img.shape[2:])
        >>> representation_tensor = portilla_simoncelli_model(img)
        >>> representation_tensor.shape
        torch.Size([1, 1, 1046])
        >>> limited_representation_tensor = portilla_simoncelli_model.remove_scales(
        ...     representation_tensor, scales_to_keep=[0]
        ... )
        >>> limited_representation_tensor.shape
        torch.Size([1, 1, 261])
        """
        # this is necessary because object is the dtype of
        # self._representation_scales
        scales_to_keep = np.array(scales_to_keep, dtype=object)
        # np.isin returns a 1d boolean array of the same shape as
        # self._representation_scales with True at each location where that
        # value appears in scales_to_keep. where then converts this boolean
        # array into indices
        ind = np.where(np.isin(self._representation_scales, scales_to_keep))[0]
        ind = torch.from_numpy(ind).to(representation_tensor.device)
        return representation_tensor.index_select(-1, ind)

    def convert_to_tensor(self, representation_dict: OrderedDict) -> Tensor:
        r"""
        Convert dictionary of statistics to a tensor.

        The output has shape (batch, channel, n_statistics), flattening and
        concatenating across all statistic classes. The dictionary representation
        may be easier to make sense of.

        Parameters
        ----------
        representation_dict
             Dictionary of representation.

        Returns
        -------
        rep
            3d tensor of statistics.

        See Also
        --------
        convert_to_dict
            Convert tensor representation to dictionary.

        Examples
        --------
        >>> import plenoptic as po
        >>> import torch
        >>> img = po.data.curie()
        >>> portilla_simoncelli_model = po.simul.PortillaSimoncelli(img.shape[2:])
        >>> representation_tensor = portilla_simoncelli_model(img)
        >>> representation_dict = portilla_simoncelli_model.convert_to_dict(
        ...     representation_tensor
        ... )
        >>> representation_tensor_new = portilla_simoncelli_model.convert_to_tensor(
        ...     representation_dict
        ... )
        >>> torch.equal(representation_tensor, representation_tensor_new)
        True
        """
        rep = einops.pack(list(representation_dict.values()), "b c *")[0]
        # then get rid of all the nans / unnecessary stats
        return rep.index_select(-1, self._necessary_stats_mask)

    def convert_to_dict(self, representation_tensor: Tensor) -> OrderedDict:
        """
        Convert tensor of statistics to a dictionary.

        While the tensor representation is required by plenoptic's synthesis
        objects, the dictionary representation is easier to manually inspect.

        This dictionary will contain NaNs in its values: these are placeholders
        for the redundant statistics.

        Parameters
        ----------
        representation_tensor
            3d tensor of statistics.

        Returns
        -------
        rep
            Dictionary of representation, with informative keys.

        Raises
        ------
        ValueError
            If ``representation_tensor`` has an unexpected number of elements. This can
            happen if some elements were manually removed from
            ``representation_tensor``, if a non-``None`` value was passed to ``forward``
            when computing it, or if it was computed using a different instantiation of
            the model.

        See Also
        --------
        convert_to_tensor:
            Convert dictionary representation to tensor.

        Examples
        --------
        >>> import plenoptic as po
        >>> img = po.data.curie()
        >>> portilla_simoncelli_model = po.simul.PortillaSimoncelli(
        ...     img.shape[2:], n_scales=3
        ... )
        >>> representation_tensor = portilla_simoncelli_model(img)
        >>> representation_dict = portilla_simoncelli_model.convert_to_dict(
        ...     representation_tensor
        ... )
        >>> # We will go through and examine each of these keys individually
        >>> # Shape is (batch, channel, 6): first four moments plus min and max
        >>> # of input image
        >>> representation_dict["pixel_statistics"].shape
        torch.Size([1, 1, 6])
        >>> # Shape is (batch, channel, spatial_corr_width, spatial_corr_width,
        >>> # n_orientations, n_scales)
        >>> representation_dict["auto_correlation_magnitude"].shape
        torch.Size([1, 1, 9, 9, 4, 3])
        >>> # Shape is (batch, channel, n_scales+1)
        >>> representation_dict["skew_reconstructed"].shape
        torch.Size([1, 1, 4])
        >>> # Shape is (batch, channel, n_scales+1)
        >>> representation_dict["kurtosis_reconstructed"].shape
        torch.Size([1, 1, 4])
        >>> # Shape is (batch, channel, spatial_corr_width, spatial_corr_width,
        >>> # n_scales+1)
        >>> representation_dict["auto_correlation_reconstructed"].shape
        torch.Size([1, 1, 9, 9, 4])
        >>> # Shape is (batch, channel, n_scales+1)
        >>> representation_dict["std_reconstructed"].shape
        torch.Size([1, 1, 4])
        >>> # Shape is (batch, channel, n_orientations, n_orientations, n_scales)
        >>> representation_dict["cross_orientation_correlation_magnitude"].shape
        torch.Size([1, 1, 4, 4, 3])
        >>> # Shape is (batch, channel, n_orientations, n_scales)
        >>> representation_dict["magnitude_std"].shape
        torch.Size([1, 1, 4, 3])
        >>> # Shape is (batch, channel, n_orientations, n_orientations, n_scales-1)
        >>> representation_dict["cross_scale_correlation_magnitude"].shape
        torch.Size([1, 1, 4, 4, 2])
        >>> # Shape is (batch, channel, n_orientations, 2*n_orientations, n_scales-1)
        >>> representation_dict["cross_scale_correlation_real"].shape
        torch.Size([1, 1, 4, 8, 2])
        >>> # Shape is (batch, channel, 1)
        >>> representation_dict["var_highpass_residual"].shape
        torch.Size([1, 1, 1])
        """
        if representation_tensor.shape[-1] != len(self._representation_scales):
            raise ValueError(
                "representation tensor is the wrong length (expected"
                f" {len(self._representation_scales)} but got"
                f" {representation_tensor.shape[-1]})! Did you remove some of"
                " the scales? (i.e., by setting scales in the forward pass)?"
                " convert_to_dict does not support such tensors."
            )

        rep = self._necessary_stats_dict.copy()
        n_filled = 0
        for k, v in rep.items():
            # each statistic is a tensor with batch and channel dimensions as
            # found in representation_tensor and all the other dimensions
            # determined by the values in necessary_stats_dict.
            shape = (*representation_tensor.shape[:2], *v.shape)
            new_v = torch.nan * torch.ones(
                shape,
                dtype=representation_tensor.dtype,
                device=representation_tensor.device,
            )
            # v.sum() gives the number of necessary elements from this stat
            this_stat_vec = representation_tensor[..., n_filled : n_filled + v.sum()]
            # use boolean indexing to put the values from new_stat_vec in the
            # appropriate place
            new_v[..., v] = this_stat_vec
            rep[k] = new_v
            n_filled += v.sum()
        return rep

    def _compute_pyr_coeffs(
        self, image: Tensor
    ) -> tuple[OrderedDict, list[Tensor], Tensor, Tensor]:
        """
        Compute pyramid coefficients of image.

        Note that the residual lowpass has been demeaned independently for each
        batch and channel (and this is true of the lowpass returned separately
        as well as the one included in pyr_coeffs_dict).

        Parameters
        ----------
        image
            4d tensor of shape (batch, channel, height, width) containing the
            image.

        Returns
        -------
        pyr_coeffs_dict
            OrderedDict of containing all pyramid coefficients.
        pyr_coeffs
            List of length n_scales, containing 5d tensors of shape (batch,
            channel, n_orientations, height, width) containing the complex-valued
            oriented bands (note that height and width shrink by half on each
            scale). This excludes the residual highpass and lowpass bands.
        highpass
            The residual highpass as a real-valued 4d tensor (batch, channel,
            height, width).
        lowpass
            The residual lowpass as a real-valued 4d tensor (batch, channel,
            height, width). This tensor has been demeaned (independently for
            each batch and channel).
        """
        pyr_coeffs = self._pyr.forward(image)
        # separate out the residuals and demean the residual lowpass
        lowpass = pyr_coeffs["residual_lowpass"]
        lowpass = lowpass - lowpass.mean(dim=(-2, -1), keepdim=True)
        pyr_coeffs["residual_lowpass"] = lowpass
        highpass = pyr_coeffs["residual_highpass"]

        # This is a list of tensors, one for each scale, where each tensor is
        # of shape (batch, channel, n_orientations, height, width) (note that
        # height and width halves on each scale)
        coeffs_list = [
            torch.stack([pyr_coeffs[(i, j)] for j in range(self.n_orientations)], 2)
            for i in range(self.n_scales)
        ]
        return pyr_coeffs, coeffs_list, highpass, lowpass

    @staticmethod
    def _compute_pixel_stats(image: Tensor) -> Tensor:
        """
        Compute the PS pixel stats: first four moments, min, and max.

        Parameters
        ----------
        image
            4d tensor of shape (batch, channel, height, width) containing input
            image. Stats are computed independently for each batch and channel.

        Returns
        -------
        pixel_stats
            3d tensor of shape (batch, channel, 6) containing the mean,
            variance, skew, kurtosis, minimum pixel value, and maximum pixel
            value (in that order).
        """  # numpydoc ignore=ES01
        mean = torch.mean(image, dim=(-2, -1), keepdim=True)
        # we use torch.var instead of plenoptic.tools.variance, because our
        # variance is the uncorrected (or sample) variance and we want the
        # corrected one here.
        var = torch.var(image, dim=(-2, -1))
        skew = stats.skew(image, mean=mean, var=var, dim=[-2, -1])
        kurtosis = stats.kurtosis(image, mean=mean, var=var, dim=[-2, -1])
        # can't compute min/max over two dims simultaneously with
        # torch.min/max, so use einops
        img_min = einops.reduce(image, "b c h w -> b c", "min")
        img_max = einops.reduce(image, "b c h w -> b c", "max")
        # mean needed to be unflattened to be used by skew and kurtosis
        # correctly, but we'll want it to be flattened like this in the final
        # representation tensor
        return einops.pack([mean, var, skew, kurtosis, img_min, img_max], "b c *")[0]

    @staticmethod
    def _compute_intermediate_representations(
        pyr_coeffs: Tensor,
    ) -> tuple[list[Tensor], list[Tensor]]:
        """
        Compute useful intermediate representations.

        These representations are:
          1) demeaned magnitude of the pyramid coefficients,
          2) real part of the pyramid coefficients

        These two are used in computing some of the texture representation.

        Parameters
        ----------
        pyr_coeffs
            Complex steerable pyramid coefficients (without residuals), as list
            of length n_scales, containing 5d tensors of shape (batch, channel,
            n_orientations, height, width).

        Returns
        -------
        magnitude_pyr_coeffs
           List of length n_scales, containing 5d tensors of shape (batch,
           channel, n_orientations, height, width) (same as ``pyr_coeffs``),
           containing the demeaned magnitude of the steerable pyramid
           coefficients (i.e., ``coeffs.abs() - coeffs.abs().mean((-2, -1))``).
        real_pyr_coeffs :
           List of length n_scales, containing 5d tensors of shape (batch,
           channel, n_orientations, height, width) (same as ``pyr_coeffs``),
           containing the real components of the coefficients (i.e.
           ``coeffs.real``).
        """
        magnitude_pyr_coeffs = [coeff.abs() for coeff in pyr_coeffs]
        magnitude_means = [
            mag.mean((-2, -1), keepdim=True) for mag in magnitude_pyr_coeffs
        ]
        magnitude_pyr_coeffs = [
            mag - mn for mag, mn in zip(magnitude_pyr_coeffs, magnitude_means)
        ]
        real_pyr_coeffs = [coeff.real for coeff in pyr_coeffs]
        return magnitude_pyr_coeffs, real_pyr_coeffs

    def _reconstruct_lowpass_at_each_scale(
        self, pyr_coeffs_dict: OrderedDict
    ) -> list[Tensor]:
        """
        Reconstruct the lowpass unoriented image at each scale.

        The autocorrelation, standard deviation, skew, and kurtosis of each of
        these images is part of the texture representation.

        Parameters
        ----------
        pyr_coeffs_dict
            Dictionary containing the steerable pyramid coefficients, with the
            lowpass residual demeaned.

        Returns
        -------
        reconstructed_images
            List of length n_scales+1 containing the reconstructed unoriented
            image at each scale, from fine to coarse. The final image is
            reconstructed just from the residual lowpass image. Each is a 4d
            tensor, this is a list because they are all different heights and
            widths.
        """
        reconstructed_images = [
            self._pyr.recon_pyr(pyr_coeffs_dict, levels=["residual_lowpass"])
        ]
        # go through scales backwards
        for lev in range(self.n_scales - 1, -1, -1):
            recon = self._pyr.recon_pyr(pyr_coeffs_dict, levels=[lev])
            reconstructed_images.append(recon + reconstructed_images[-1])
        # now downsample as necessary, so that these end up the same size as
        # their corresponding coefficients. We multiply by the factor of 4 here
        # in order to approximately equalize the steerable pyramid coefficient
        # values across scales. This could also be handled by making the
        # pyramid tight frame
        reconstructed_images[:-1] = [
            signal.shrink(r, 2 ** (self.n_scales - i)) * 4 ** (self.n_scales - i)
            for i, r in enumerate(reconstructed_images[:-1])
        ]
        return reconstructed_images

    def _compute_autocorr(self, coeffs_list: list[Tensor]) -> tuple[Tensor, Tensor]:
        """
        Compute the autocorrelation of some statistics.

        Parameters
        ----------
        coeffs_list
            List (of length s) of tensors of shape (batch, channel, *, height,
            width), where * is zero or one additional dimensions. Intended use
            case: ``magnitude_pyr_coeffs`` (which is list of length ``n_scales`` of 5d
            tensors, with * containing ``n_orientations``) or ``reconstructed_images``
            (which is a list of length ``n_scales+1`` of 4d tensors).

        Returns
        -------
        autocorrs
            Tensor of shape (batch, channel, spatial_corr_width,
            spatial_corr_width, *, s) containing the autocorrelation (up to
            distance ``spatial_corr_width//2``) of each element in
            ``coeffs_list``, computed independently over all but the final two
            dimensions.
        vars
            3d Tensor of shape (batch, channel, *, s) containing the variance
            of each element in ``coeffs_list``, computed independently over all
            but the final two dimensions.

        Raises
        ------
        ValueError
            If ``coeffs_list`` contains tensors that have other than 4 or 5 dimensions.
        """  # numpydoc ignore=ES01
        if coeffs_list[0].ndim == 5:
            dims = "o"
        elif coeffs_list[0].ndim == 4:
            dims = ""
        else:
            raise ValueError(
                "coeffs_list must contain tensors of either 4 or 5 dimensions!"
            )
        acs = [signal.autocorrelation(coeff) for coeff in coeffs_list]
        var = [signal.center_crop(ac, 1) for ac in acs]
        acs = [ac / v for ac, v in zip(acs, var)]
        var = einops.rearrange(var, f"s b c {dims} 1 1 -> b c {dims} s")
        acs = [signal.center_crop(ac, self.spatial_corr_width) for ac in acs]
        acs = torch.stack(acs, 2)
        return einops.rearrange(acs, f"b c s {dims} a1 a2 -> b c a1 a2 {dims} s"), var

    @staticmethod
    def _compute_skew_kurtosis_recon(
        reconstructed_images: list[Tensor], var_recon: Tensor, img_var: Tensor
    ) -> tuple[Tensor, Tensor]:
        """
        Compute the skew and kurtosis of each lowpass reconstructed image.

        For each scale, if the ratio of its variance to the original image's
        pixel variance is below a threshold of
        ``torch.finfo(img_var.dtype).resolution`` (``1e-6`` for ``float32``,
        ``1e-15`` for ``float64``), skew and kurtosis are assigned default
        values of ``0`` or ``3``, respectively.

        Parameters
        ----------
        reconstructed_images
            List of length ``n_scales+1`` containing the reconstructed unoriented
            image at each scale, from fine to coarse. The final image is
            reconstructed just from the residual lowpass image.
        var_recon
            Tensor of shape (batch, channel, n_scales+1) containing the
            variance of each tensor in reconstruced_images.
        img_var
            Tensor of shape (batch, channel) containing the pixel variance
            (from ``pixel_stats`` tensor).

        Returns
        -------
        skew_recon, kurtosis_recon
            Tensors of shape (batch, channel, n_scales+1) containing the skew
            and kurtosis, respectively, of each tensor in
            ``reconstructed_images``.
        """
        skew_recon = [
            stats.skew(im, mean=0, var=var_recon[..., i], dim=[-2, -1])
            for i, im in enumerate(reconstructed_images)
        ]
        skew_recon = torch.stack(skew_recon, -1)
        kurtosis_recon = [
            stats.kurtosis(im, mean=0, var=var_recon[..., i], dim=[-2, -1])
            for i, im in enumerate(reconstructed_images)
        ]
        kurtosis_recon = torch.stack(kurtosis_recon, -1)
        skew_default = torch.zeros_like(skew_recon)
        kurtosis_default = 3 * torch.ones_like(kurtosis_recon)
        # if this variance ratio is too small, then use the default values
        # instead. unsqueeze is used here because var_recon is shape (batch,
        # channel, scales+1), whereas img_var is just (batch, channel)
        res = torch.finfo(img_var.dtype).resolution
        unstable_locs = var_recon / img_var.unsqueeze(-1) < res
        skew_recon = torch.where(unstable_locs, skew_default, skew_recon)
        kurtosis_recon = torch.where(unstable_locs, kurtosis_default, kurtosis_recon)
        return skew_recon, kurtosis_recon

    def _compute_cross_correlation(
        self,
        coeffs_tensor: list[Tensor],
        coeffs_tensor_other: list[Tensor],
        coeffs_var: None | Tensor = None,
        coeffs_other_var: None | Tensor = None,
    ) -> Tensor:
        """
        Compute cross-correlations.

        Parameters
        ----------
        coeffs_tensor, coeffs_tensor_other
            The two lists of length scales, each containing 5d tensors of shape
            (batch, channel, n_orientations, height, width) to be correlated.
        coeffs_var, coeffs_other_var
            Two optional tensors containing the variances of coeffs_tensor and
            coeffs_tensor_other, respectively, in case they've already been computed.
            Should be of shape (batch, channel, n_orientations, n_scales). Used to
            normalize the covariances into cross-correlations.

        Returns
        -------
        cross_corrs
            Tensor of shape (batch, channel, n_orientations, n_orientations,
            scales) containing the cross-correlations at each
            scale.
        """  # numpydoc ignore=ES01
        covars = []
        for i, (coeff, coeff_other) in enumerate(
            zip(coeffs_tensor, coeffs_tensor_other)
        ):
            # precompute this, which we'll use for normalization
            numel = torch.mul(*coeff.shape[-2:])
            # compute the covariance
            covar = einops.einsum(
                coeff, coeff_other, "b c o1 h w, b c o2 h w -> b c o1 o2"
            )
            covar = covar / numel
            # Then normalize it to get the Pearson product-moment correlation
            # coefficient, see
            # https://numpy.org/doc/stable/reference/generated/numpy.corrcoef.html.
            if coeffs_var is None:
                # First, compute the variances of each coeff
                coeff_var = einops.einsum(
                    coeff, coeff, "b c o1 h w, b c o1 h w -> b c o1"
                )
                coeff_var = coeff_var / numel
            else:
                coeff_var = coeffs_var[..., i]
            if coeffs_other_var is None:
                # First, compute the variances of each coeff
                coeff_other_var = einops.einsum(
                    coeff_other, coeff_other, "b c o1 h w, b c o1 h w -> b c o1"
                )
                coeff_other_var = coeff_other_var / numel
            else:
                coeff_other_var = coeffs_other_var[..., i]
            # Then compute the outer product of those variances.
            var_outer_prod = einops.einsum(
                coeff_var, coeff_other_var, "b c o1, b c o2 -> b c o1 o2"
            )
            # And the sqrt of this is what we use to normalize the covariance
            # into the cross-correlation
            covars.append(covar / var_outer_prod.sqrt())
        return torch.stack(covars, -1)

    @staticmethod
    def _double_phase_pyr_coeffs(
        pyr_coeffs: list[Tensor],
    ) -> tuple[list[Tensor], list[Tensor]]:
        """
        Upsample and double the phase of pyramid coefficients.

        This is trick is key to correctly computing the correlation between
        coefficients at different spatial scales.

        Parameters
        ----------
        pyr_coeffs
            Complex steerable pyramid coefficients (without residuals), as list
            of length n_scales, containing 5d tensors of shape (batch, channel,
            n_orientations, height, width).

        Returns
        -------
        doubled_phase_mags
            The demeaned magnitude (i.e., pyr_coeffs.abs()) of each upsampled
            double-phased coefficient. List of length n_scales-1 containing
            tensors of same shape the input (the finest scale has been
            removed).
        doubled_phase_separate
            The real and imaginary parts of each double-phased coefficient.
            List of length n_scales-1, containing tensors of shape (batch,
            channel, 2*n_orientations, height, width), with the real component
            found at the same orientation index as the input, and the imaginary
            at orientation+self.n_orientations. (The finest scale has been
            removed).
        """
        doubled_phase_mags = []
        doubled_phase_sep = []
        # don't do this for the finest scale
        for coeff in pyr_coeffs[1:]:
            # We divide by the factor of 4 here in order to approximately
            # equalize the steerable pyramid coefficient values across scales.
            # This could also be handled by making the pyramid tight frame
            doubled_phase = signal.expand(coeff, 2) / 4.0
            doubled_phase = signal.modulate_phase(doubled_phase, 2)
            doubled_phase_mag = doubled_phase.abs()
            doubled_phase_mag = doubled_phase_mag - doubled_phase_mag.mean(
                (-2, -1), keepdim=True
            )
            doubled_phase_mags.append(doubled_phase_mag)
            doubled_phase_sep.append(
                einops.pack([doubled_phase.real, doubled_phase.imag], "b c * h w")[0]
            )
        return doubled_phase_mags, doubled_phase_sep

    def plot_representation(
        self,
        data: Tensor,
        ax: plt.Axes | None = None,
        figsize: tuple[float, float] | None = None,
        ylim: tuple[float, float] | Literal[False] | None = None,
        batch_idx: int = 0,
        title: str | None = None,
    ) -> tuple[plt.Figure, list[plt.Axes]]:
        r"""
        Plot the representation in a human viewable format.

        We plot the representation as stem plots with data separated out by
        statistic type.

        This plots the representation of a single batch and averages over all
        channels in the representation.

        We create the following axes:

        - pixels+var_highpass: marginal pixel statistics (first four moments,
          min, max) and variance of the residual highpass.

        - std+skew+kurtosis recon: the standard deviation, skew, and kurtosis
          of the reconstructed lowpass image at each scale

        - magnitude_std: the standard deviation of the steerable pyramid
          coefficient magnitudes at each orientation and scale.

        - auto_correlation_reconstructed: the auto-correlation of the
          reconstructed lowpass image at each scale (summarized using Euclidean
          norm).

        - auto_correlation_magnitude: the auto-correlation of the pyramid
          coefficient magnitudes at each scale and orientation (summarized
          using Euclidean norm).

        - cross_orientation_correlation_magnitude: the cross-correlations
          between each orientation at each scale (summarized using Euclidean
          norm)

        If ``self.n_scales > 1``, we also have combination of the following, where
        all cross-correlations are summarized using Euclidean norm over the
        channel dimension:

        - cross_scale_correlation_magnitude: the cross-correlations between the
          pyramid coefficient magnitude at one scale and the same orientation
          at the next-coarsest scale.

        - cross_scale_correlation_real: the cross-correlations between the real
          component of the pyramid coefficients and the real and imaginary
          components (at the same orientation) at the next-coarsest scale.

        Parameters
        ----------
        data
            The data to show on the plot. Else, should look like the output of
            ``self.forward(img)``, with the exact same structure (e.g., as
            returned by ``metamer.representation_error()`` or another instance
            of this class).
        ax
            Axes where we will plot the data. If a ``plt.Axes`` instance, will
            subdivide into 6 or 8 new axes (depending on self.n_scales). If
            None, we create a new figure.
        figsize
            The size of the figure to create. Must be ``None`` if ax is not ``None``. If
            both figsize and ax are ``None``, then we set ``figsize=(15, 15)``.
        ylim
            If not None, the y-limits to use for this plot. If None, we use the
            default, slightly adjusted so that the minimum is 0. If False, do not
            change y-limits.
        batch_idx
            Which index to take from the batch dimension (the first one).
        title
            Title for the plot.

        Returns
        -------
        fig
            Figure containing the plot.
        axes
            List of 6 or 8 axes containing the plot (depending on ``self.n_scales``).

        Raises
        ------
        ValueError
            If both ``figsize`` and ``ax`` are not ``None``.

        Examples
        --------
        .. plot::

          >>> import plenoptic as po
          >>> img = po.data.curie()
          >>> portilla_simoncelli_model = po.simul.PortillaSimoncelli(img.shape[2:])
          >>> representation_tensor = portilla_simoncelli_model(img)
          >>> fig, axes = portilla_simoncelli_model.plot_representation(
          ...     representation_tensor, figsize=(13, 6)
          ... )
        """
        if ax is None and figsize is None:
            figsize = (15, 15)
        elif ax is not None and figsize is not None:
            raise ValueError("figsize can't be set if ax is not None")
        # pick the batch_idx we want (but keep the data 3d), and average over
        # channels (but keep the data 3d). We keep data 3d because
        # convert_to_dict relies on it.
        data = data[batch_idx].unsqueeze(0).mean(1, keepdim=True)
        # each of these values should now be a 3d tensor with 1 element in each
        # of the first two dims
        rep = {k: v[0, 0] for k, v in self.convert_to_dict(data).items()}
        data = self._representation_for_plotting(rep)

        # Determine plot grid layout
        if self.n_scales != 1:
            n_rows = 3
            n_cols = int(np.ceil(len(data) / n_rows))
        else:
            # then we don't have any cross-scale correlations, so fewer axes.
            n_rows = 2
            n_cols = int(np.ceil(len(data) / n_rows))

        # Set up grid spec
        if ax is None:
            # we add 2 to order because we're adding one to get the
            # number of orientations and then another one to add an
            # extra column for the mean luminance plot
            fig = plt.figure(figsize=figsize)
            gs = mpl.gridspec.GridSpec(n_rows, n_cols, fig)
        else:
            # want to make sure the axis we're taking over is basically invisible.
            ax = clean_up_axes(
                ax, False, ["top", "right", "bottom", "left"], ["x", "y"]
            )
            gs = ax.get_subplotspec().subgridspec(n_rows, n_cols)
            fig = ax.figure

        # plot data
        axes = []
        for i, (k, v) in enumerate(data.items()):
            ax = fig.add_subplot(gs[i // n_cols, i % n_cols])
            ax = clean_stem_plot(to_numpy(v).flatten(), ax, k, ylim=ylim)
            axes.append(ax)

        if title is not None:
            fig.suptitle(title)

        return fig, axes

    def _representation_for_plotting(self, rep: OrderedDict) -> OrderedDict:
        r"""
        Convert into a representation that is more convenient for plotting.

        Intended as a helper function for plot_representation.

        Parameters
        ----------
        rep
            Dictionary of representation, with informative keys.

        Returns
        -------
        plot_rep
            Dictionary of representation summarized for plotting.

        Raises
        ------
        ValueError
            If the tensors in ``rep`` looks like they have more than one batch
            or channel. Should select or average over those dimensions.
        ValueError
            If ``rep`` contains unexpected keys.
        """
        if rep["skew_reconstructed"].ndim > 1:
            raise ValueError(
                "Currently, only know how to plot single batch and channel at"
                " a time! Select and/or average over those dimensions"
            )
        data = OrderedDict()
        data["pixels+var_highpass"] = torch.cat(
            [rep.pop("pixel_statistics"), rep.pop("var_highpass_residual")]
        )
        data["std+skew+kurtosis recon"] = torch.cat(
            (
                rep.pop("std_reconstructed"),
                rep.pop("skew_reconstructed"),
                rep.pop("kurtosis_reconstructed"),
            )
        )

        data["magnitude_std"] = rep.pop("magnitude_std")

        # want to plot these in a specific order
        all_keys = [
            "auto_correlation_reconstructed",
            "auto_correlation_magnitude",
            "cross_orientation_correlation_magnitude",
            "cross_scale_correlation_magnitude",
            "cross_scale_correlation_real",
        ]
        if set(rep.keys()) != set(all_keys):
            raise ValueError("representation has unexpected keys!")
        for k in all_keys:
            # if we only have one scale, no cross-scale stats
            if k.startswith("cross_scale") and self.n_scales == 1:
                continue
            # we compute L2 norm manually, since there are NaNs (marking
            # redundant stats)
            data[k] = rep[k].pow(2).nansum((0, 1)).sqrt().flatten()

        return data

    def update_plot(
        self,
        axes: list[plt.Axes],
        data: Tensor,
        batch_idx: int = 0,
    ) -> list[plt.Artist]:
        r"""
        Update the information in our representation plot.

        This is used for creating an animation of the representation
        over time. In order to create the animation, we need to know how
        to update the matplotlib Artists, and this provides a simple way
        of doing that. It relies on the fact that we've used
        ``plot_representation`` to create the plots we want to update
        and so know that they're stem plots.

        We take the axes containing the representation information (note that
        this is probably a subset of the total number of axes in the figure, if
        we're showing other information, as done by ``Metamer.animate``), grab
        the representation from plotting and, since these are both lists,
        iterate through them, updating them to the values in ``data`` as we go.

        In order for this to be used by ``FuncAnimation``, we need to
        return Artists, so we return a list of the relevant artists, the
        ``markerline`` and ``stemlines`` from the ``StemContainer``.

        Currently, this averages over all channels in the representation.

        Parameters
        ----------
        axes
            A list of axes to update. We assume that these are the axes
            created by ``plot_representation`` and so contain stem plots
            in the correct order.
        data
            The data to show on the plot. Else, should look like the output of
            ``self.forward(img)``, with the exact same structure (e.g., as
            returned by ``metamer.representation_error()`` or another instance
            of this class).
        batch_idx
            Which index to take from the batch dimension (the first one).

        Returns
        -------
        stem_artists
            A list of the artists used to update the information on the
            stem plots.

        Examples
        --------
        This method is meant to be used by animation functions, so users won't
        typically use this directly.

        >>> import plenoptic as po
        >>> img = po.data.curie()
        >>> portilla_simoncelli_model = po.simul.PortillaSimoncelli(img.shape[2:])
        >>> representation_tensor = portilla_simoncelli_model.forward(img)
        >>> fig, axes = portilla_simoncelli_model.plot_representation(
        ...     representation_tensor
        ... )
        >>> new_img = po.data.einstein()
        >>> new_representation_tensor = portilla_simoncelli_model.forward(new_img)
        >>> stem_artists = portilla_simoncelli_model.update_plot(
        ...     axes, new_representation_tensor
        ... )
        """
        stem_artists = []
        axes = [ax for ax in axes if len(ax.containers) == 1]
        # pick the batch_idx we want (but keep the data 3d), and average over
        # channels (but keep the data 3d). We keep data 3d because
        # convert_to_dict relies on it.
        data = data[batch_idx].unsqueeze(0).mean(1, keepdim=True)
        # each of these values should now be a 3d tensor with 1 element in each
        # of the first two dims
        rep = {k: v[0, 0] for k, v in self.convert_to_dict(data).items()}
        rep = self._representation_for_plotting(rep)
        for ax, d in zip(axes, rep.values()):
            if isinstance(d, dict):
                vals = np.array([to_numpy(dd) for dd in d.values()])
            else:
                vals = to_numpy(d.flatten())

            sc = update_stem(ax.containers[0], vals)
            stem_artists.extend([sc.markerline, sc.stemlines])
        return stem_artists
